from typing import Optional

from validator_collection import validators

from highcharts_core import constants
from highcharts_core.decorators import class_sensitive
from highcharts_core.metaclasses import HighchartsMeta


class PopupLanguageOptions(HighchartsMeta):
    """Translations for all field names used in popup."""

    def __init__(self, **kwargs):
        self._add_button = None
        self._algorithm = None
        self._arrow_infinity_line = None
        self._arrow_ray = None
        self._arrow_segment = None
        self._average = None
        self._background = None
        self._background_color = None
        self._background_colors = None
        self._border_color = None
        self._border_radius = None
        self._border_width = None
        self._bottom_band = None
        self._circle = None
        self._clear_filter = None
        self._color = None
        self._connector = None
        self._crooked3 = None
        self._crooked5 = None
        self._crosshairX = None
        self._crosshairY = None
        self._decimals = None
        self._deviation = None
        self._edit_button = None
        self._elliott3 = None
        self._elliott5 = None
        self._ellipse = None
        self._factor = None
        self._fast_avg_period = None
        self._fibonacci = None
        self._fibonacci_time_zones = None
        self._fill = None
        self._flags = None
        self._font_size = None
        self._format = None
        self._height = None
        self._high_index = None
        self._horizontal_line = None
        self._increment = None
        self._index = None
        self._infinity_line = None
        self._initial_acceleration_factor = None
        self._inner_background = None
        self._label = None
        self._label_options = None
        self._labels = None
        self._line = None
        self._lines = None
        self._long_period = None
        self._low_index = None
        self._max_acceleration_factor = None
        self._measure = None
        self._measure_x = None
        self._measure_xy = None
        self._measure_y = None
        self._multiplier = None
        self._multiplier_atr = None
        self._name = None
        self._no_filter_match = None
        self._outer_background = None
        self._padding = None
        self._parallel_channel = None
        self._period = None
        self._period_atr = None
        self._periods = None
        self._period_senkou_span_b = None
        self._period_tenkan = None
        self._pitchfork = None
        self._ranges = None
        self._ray = None
        self._rectangle = None
        self._remove_button = None
        self._save_button = None
        self._search_indicators = None
        self._segment = None
        self._series = None
        self._shape_options = None
        self._shapes = None
        self._short_period = None
        self._signal_period = None
        self._simple_shapes = None
        self._slow_avg_period = None
        self._standard_deviation = None
        self._stroke = None
        self._stroke_width = None
        self._style = None
        self._time_cycles = None
        self._title = None
        self._top_band = None
        self._tunnel = None
        self._type_options = None
        self._vertical_arrow = None
        self._vertical_counter = None
        self._vertical_label = None
        self._vertical_line = None
        self._volume = None
        self._x_axis_unit = None

        self.add_button = kwargs.get('add_button', None)
        self.algorithm = kwargs.get('algorithm', None)
        self.arrow_infinity_line = kwargs.get('arrow_infinity_line', None)
        self.arrow_ray = kwargs.get('arrow_ray', None)
        self.arrow_segment = kwargs.get('arrow_segment', None)
        self.average = kwargs.get('average', None)
        self.background = kwargs.get('background', None)
        self.background_color = kwargs.get('background_color', None)
        self.background_colors = kwargs.get('background_colors', None)
        self.border_color = kwargs.get('border_color', None)
        self.border_radius = kwargs.get('border_radius', None)
        self.border_width = kwargs.get('border_width', None)
        self.bottom_band = kwargs.get('bottom_band', None)
        self.circle = kwargs.get('circle', None)
        self.clear_filter = kwargs.get('clear_filter', None)
        self.color = kwargs.get('color', None)
        self.connector = kwargs.get('connector', None)
        self.crooked3 = kwargs.get('crooked3', None)
        self.crooked5 = kwargs.get('crooked5', None)
        self.crosshairX = kwargs.get('crosshairX', None)
        self.crosshairY = kwargs.get('crosshairY', None)
        self.decimals = kwargs.get('decimals', None)
        self.deviation = kwargs.get('deviation', None)
        self.edit_button = kwargs.get('edit_button', None)
        self.elliott3 = kwargs.get('elliott3', None)
        self.elliott5 = kwargs.get('elliott5', None)
        self.ellipse = kwargs.get('ellipse', None)
        self.factor = kwargs.get('factor', None)
        self.fast_avg_period = kwargs.get('fast_avg_period', None)
        self.fibonacci = kwargs.get('fibonacci', None)
        self.fibonacci_time_zones = kwargs.get('fibonacci_time_zones', None)
        self.fill = kwargs.get('fill', None)
        self.flags = kwargs.get('flags', None)
        self.font_size = kwargs.get('font_size', None)
        self.format = kwargs.get('format', None)
        self.height = kwargs.get('height', None)
        self.high_index = kwargs.get('high_index', None)
        self.horizontal_line = kwargs.get('horizontal_line', None)
        self.increment = kwargs.get('increment', None)
        self.index = kwargs.get('index', None)
        self.infinity_line = kwargs.get('infinity_line', None)
        self.initial_acceleration_factor = kwargs.get('initial_acceleration_factor', None)
        self.inner_background = kwargs.get('inner_background', None)
        self.label = kwargs.get('label', None)
        self.label_options = kwargs.get('label_options', None)
        self.labels = kwargs.get('labels', None)
        self.line = kwargs.get('line', None)
        self.lines = kwargs.get('lines', None)
        self.long_period = kwargs.get('long_period', None)
        self.low_index = kwargs.get('low_index', None)
        self.max_acceleration_factor = kwargs.get('max_acceleration_factor', None)
        self.measure = kwargs.get('measure', None)
        self.measure_x = kwargs.get('measure_x', None)
        self.measure_xy = kwargs.get('measure_xy', None)
        self.measure_y = kwargs.get('measure_y', None)
        self.multiplier = kwargs.get('multiplier', None)
        self.multiplier_atr = kwargs.get('multiplier_atr', None)
        self.name = kwargs.get('name', None)
        self.no_filter_match = kwargs.get('no_filter_match', None)
        self.outer_background = kwargs.get('outer_background', None)
        self.padding = kwargs.get('padding', None)
        self.parallel_channel = kwargs.get('parallel_channel', None)
        self.period = kwargs.get('period', None)
        self.period_atr = kwargs.get('period_atr', None)
        self.periods = kwargs.get('periods', None)
        self.period_senkou_span_b = kwargs.get('period_senkou_span_b', None)
        self.period_tenkan = kwargs.get('period_tenkan', None)
        self.pitchfork = kwargs.get('pitchfork', None)
        self.ranges = kwargs.get('ranges', None)
        self.ray = kwargs.get('ray', None)
        self.rectangle = kwargs.get('rectangle', None)
        self.remove_button = kwargs.get('remove_button', None)
        self.save_button = kwargs.get('save_button', None)
        self.search_indicators = kwargs.get('search_indicators', None)
        self.segment = kwargs.get('segment', None)
        self.series = kwargs.get('series', None)
        self.shape_options = kwargs.get('shape_options', None)
        self.shapes = kwargs.get('shapes', None)
        self.short_period = kwargs.get('short_period', None)
        self.signal_period = kwargs.get('signal_period', None)
        self.simple_shapes = kwargs.get('simple_shapes', None)
        self.slow_avg_period = kwargs.get('slow_avg_period', None)
        self.standard_deviation = kwargs.get('standard_deviation', None)
        self.stroke = kwargs.get('stroke', None)
        self.stroke_width = kwargs.get('stroke_width', None)
        self.style = kwargs.get('style', None)
        self.time_cycles = kwargs.get('time_cycles', None)
        self.title = kwargs.get('title', None)
        self.top_band = kwargs.get('top_band', None)
        self.tunnel = kwargs.get('tunnel', None)
        self.type_options = kwargs.get('type_options', None)
        self.vertical_arrow = kwargs.get('vertical_arrow', None)
        self.vertical_counter = kwargs.get('vertical_counter', None)
        self.vertical_label = kwargs.get('vertical_label', None)
        self.vertical_line = kwargs.get('vertical_line', None)
        self.volume = kwargs.get('volume', None)
        self.x_axis_unit = kwargs.get('x_axis_unit', None)

    @property
    def add_button(self) -> Optional[str]:
        """Defaults to ``'add'``.

        :rtype: :class:`str <python:str>` or :obj:`None <python:None>`
        """
        return self._add_button

    @add_button.setter
    def add_button(self, value):
        self._add_button = validators.string(value, allow_empty = True)

    @property
    def algorithm(self) -> Optional[str]:
        """Defaults to ``'Algorithm'``.

        :rtype: :class:`str <python:str>` or :obj:`None <python:None>`
        """
        return self._algorithm

    @algorithm.setter
    def algorithm(self, value):
        self._algorithm = validators.string(value, allow_empty = True)

    @property
    def arrow_infinity_line(self) -> Optional[str]:
        """Defaults to
        ``'Arrow line'``.

        :rtype: :class:`str <python:str>` or :obj:`None <python:None>`
        """
        return self._arrow_infinity_line

    @arrow_infinity_line.setter
    def arrow_infinity_line(self, value):
        self._arrow_infinity_line = validators.string(value, allow_empty = True)

    @property
    def arrow_ray(self) -> Optional[str]:
        """Defaults to ``'Arrow ray'``.

        :rtype: :class:`str <python:str>` or :obj:`None <python:None>`
        """
        return self._arrow_ray

    @arrow_ray.setter
    def arrow_ray(self, value):
        self._arrow_ray = validators.string(value, allow_empty = True)

    @property
    def arrow_segment(self) -> Optional[str]:
        """Defaults to ``'Arrow segment'``.

        :rtype: :class:`str <python:str>` or :obj:`None <python:None>`
        """
        return self._arrow_segment

    @arrow_segment.setter
    def arrow_segment(self, value):
        self._arrow_segment = validators.string(value, allow_empty = True)

    @property
    def average(self) -> Optional[str]:
        """Defaults to ``'Average'``.

        :rtype: :class:`str <python:str>` or :obj:`None <python:None>`
        """
        return self._average

    @average.setter
    def average(self, value):
        self._average = validators.string(value, allow_empty = True)

    @property
    def background(self) -> Optional[str]:
        """Defaults to ``'Background'``.

        :rtype: :class:`str <python:str>` or :obj:`None <python:None>`
        """
        return self._background

    @background.setter
    def background(self, value):
        self._background = validators.string(value, allow_empty = True)

    @property
    def background_color(self) -> Optional[str]:
        """Defaults to ``'Background color'``.

        :rtype: :class:`str <python:str>` or :obj:`None <python:None>`
        """
        return self._background_color

    @background_color.setter
    def background_color(self, value):
        self._background_color = validators.string(value, allow_empty = True)

    @property
    def background_colors(self) -> Optional[str]:
        """Defaults to
        ``'Background colors'``.

        :rtype: :class:`str <python:str>` or :obj:`None <python:None>`
        """
        return self._background_colors

    @background_colors.setter
    def background_colors(self, value):
        self._background_colors = validators.string(value, allow_empty = True)

    @property
    def border_color(self) -> Optional[str]:
        """Defaults to ``'Border color'``.

        :rtype: :class:`str <python:str>` or :obj:`None <python:None>`
        """
        return self._border_color

    @border_color.setter
    def border_color(self, value):
        self._border_color = validators.string(value, allow_empty = True)

    @property
    def border_radius(self) -> Optional[str]:
        """Defaults to ``'Border radius'``.

        :rtype: :class:`str <python:str>` or :obj:`None <python:None>`
        """
        return self._border_radius

    @border_radius.setter
    def border_radius(self, value):
        self._border_radius = validators.string(value, allow_empty = True)

    @property
    def border_width(self) -> Optional[str]:
        """Defaults to ``'Border width'``.

        :rtype: :class:`str <python:str>` or :obj:`None <python:None>`
        """
        return self._border_width

    @border_width.setter
    def border_width(self, value):
        self._border_width = validators.string(value, allow_empty = True)

    @property
    def bottom_band(self) -> Optional[str]:
        """Defaults to ``'Bottom band'``.

        :rtype: :class:`str <python:str>` or :obj:`None <python:None>`
        """
        return self._bottom_band

    @bottom_band.setter
    def bottom_band(self, value):
        self._bottom_band = validators.string(value, allow_empty = True)

    @property
    def circle(self) -> Optional[str]:
        """Defaults to ``'Circle'``.

        :rtype: :class:`str <python:str>` or :obj:`None <python:None>`
        """
        return self._circle

    @circle.setter
    def circle(self, value):
        self._circle = validators.string(value, allow_empty = True)

    @property
    def clear_filter(self) -> Optional[str]:
        """Defaults to ``'X Clear Filter'``.

        :rtype: :class:`str <python:str>` or :obj:`None <python:None>`
        """
        return self._clear_filter

    @clear_filter.setter
    def clear_filter(self, value):
        self._clear_filter = validators.string(value, allow_empty = True)

    @property
    def color(self) -> Optional[str]:
        """Defaults to ``'Color'``.

        :rtype: :class:`str <python:str>` or :obj:`None <python:None>`
        """
        return self._color

    @color.setter
    def color(self, value):
        self._color = validators.string(value, allow_empty = True)

    @property
    def connector(self) -> Optional[str]:
        """Defaults to ``'Connector'``.

        :rtype: :class:`str <python:str>` or :obj:`None <python:None>`
        """
        return self._connector

    @connector.setter
    def connector(self, value):
        self._connector = validators.string(value, allow_empty = True)

    @property
    def crooked3(self) -> Optional[str]:
        """Defaults to ``'Crooked 3 line'``.

        :rtype: :class:`str <python:str>` or :obj:`None <python:None>`
        """
        return self._crooked3

    @crooked3.setter
    def crooked3(self, value):
        self._crooked3 = validators.string(value, allow_empty = True)

    @property
    def crooked5(self) -> Optional[str]:
        """Defaults to ``'Crooked 5 line'``.

        :rtype: :class:`str <python:str>` or :obj:`None <python:None>`
        """
        return self._crooked5

    @crooked5.setter
    def crooked5(self, value):
        self._crooked5 = validators.string(value, allow_empty = True)

    @property
    def crosshairX(self) -> Optional[str]:
        """Defaults to ``'Crosshair X'``.

        :rtype: :class:`str <python:str>` or :obj:`None <python:None>`
        """
        return self._crosshairX

    @crosshairX.setter
    def crosshairX(self, value):
        self._crosshairX = validators.string(value, allow_empty = True)

    @property
    def crosshairY(self) -> Optional[str]:
        """Defaults to ``'Crosshair Y'``.

        :rtype: :class:`str <python:str>` or :obj:`None <python:None>`
        """
        return self._crosshairY

    @crosshairY.setter
    def crosshairY(self, value):
        self._crosshairY = validators.string(value, allow_empty = True)

    @property
    def decimals(self) -> Optional[str]:
        """Defaults to ``'Decimals``.

        :rtype: :class:`str <python:str>` or :obj:`None <python:None>`
        """
        return self._decimals

    @decimals.setter
    def decimals(self, value):
        self._decimals = validators.string(value, allow_empty = True)

    @property
    def deviation(self) -> Optional[str]:
        """Defaults to ``'Deviation'``.

        :rtype: :class:`str <python:str>` or :obj:`None <python:None>`
        """
        return self._deviation

    @deviation.setter
    def deviation(self, value):
        self._deviation = validators.string(value, allow_empty = True)

    @property
    def edit_button(self) -> Optional[str]:
        """Defaults to ``'edit'``.

        :rtype: :class:`str <python:str>` or :obj:`None <python:None>`
        """
        return self._edit_button

    @edit_button.setter
    def edit_button(self, value):
        self._edit_button = validators.string(value, allow_empty = True)

    @property
    def elliott3(self) -> Optional[str]:
        """Defaults to ``'Elliott 3 line'``.

        :rtype: :class:`str <python:str>` or :obj:`None <python:None>`
        """
        return self._elliott3

    @elliott3.setter
    def elliott3(self, value):
        self._elliott3 = validators.string(value, allow_empty = True)

    @property
    def elliott5(self) -> Optional[str]:
        """Defaults to ``'Elliott 5 line'``.

        :rtype: :class:`str <python:str>` or :obj:`None <python:None>`
        """
        return self._elliott5

    @elliott5.setter
    def elliott5(self, value):
        self._elliott5 = validators.string(value, allow_empty = True)

    @property
    def ellipse(self) -> Optional[str]:
        """Defaults to ``'Ellipse``.

        :rtype: :class:`str <python:str>` or :obj:`None <python:None>`
        """
        return self._ellipse

    @ellipse.setter
    def ellipse(self, value):
        self._ellipse = validators.string(value, allow_empty = True)

    @property
    def factor(self) -> Optional[str]:
        """Defaults to ``'Factor``.

        :rtype: :class:`str <python:str>` or :obj:`None <python:None>`
        """
        return self._factor

    @factor.setter
    def factor(self, value):
        self._factor = validators.string(value, allow_empty = True)

    @property
    def fast_avg_period(self) -> Optional[str]:
        """Defaults to ``'Fast average period'``.

        :rtype: :class:`str <python:str>` or :obj:`None <python:None>`
        """
        return self._fast_avg_period

    @fast_avg_period.setter
    def fast_avg_period(self, value):
        self._fast_avg_period = validators.string(value, allow_empty = True)

    @property
    def fibonacci(self) -> Optional[str]:
        """Defaults to ``'Fibonacci``.

        :rtype: :class:`str <python:str>` or :obj:`None <python:None>`
        """
        return self._fibonacci

    @fibonacci.setter
    def fibonacci(self, value):
        self._fibonacci = validators.string(value, allow_empty = True)

    @property
    def fibonacci_time_zones(self) -> Optional[str]:
        """Defaults to
        ``'Fibonacci Time Zones'``.

        :rtype: :class:`str <python:str>` or :obj:`None <python:None>`
        """
        return self._fibonacci_time_zones

    @fibonacci_time_zones.setter
    def fibonacci_time_zones(self, value):
        self._fibonacci_time_zones = validators.string(value, allow_empty = True)

    @property
    def fill(self) -> Optional[str]:
        """Defaults to ``'Fill``.

        :rtype: :class:`str <python:str>` or :obj:`None <python:None>`
        """
        return self._fill

    @fill.setter
    def fill(self, value):
        self._fill = validators.string(value, allow_empty = True)

    @property
    def flags(self) -> Optional[str]:
        """Defaults to ``'Flags``.

        :rtype: :class:`str <python:str>` or :obj:`None <python:None>`
        """
        return self._flags

    @flags.setter
    def flags(self, value):
        self._flags = validators.string(value, allow_empty = True)

    @property
    def font_size(self) -> Optional[str]:
        """Defaults to ``'Font size'``.

        :rtype: :class:`str <python:str>` or :obj:`None <python:None>`
        """
        return self._font_size

    @font_size.setter
    def font_size(self, value):
        self._font_size = validators.string(value, allow_empty = True)

    @property
    def format(self) -> Optional[str]:
        """Defaults to ``'Text'``.

        :rtype: :class:`str <python:str>` or :obj:`None <python:None>`
        """
        return self._format

    @format.setter
    def format(self, value):
        self._format = validators.string(value, allow_empty = True)

    @property
    def height(self) -> Optional[str]:
        """Defaults to ``'Height``.

        :rtype: :class:`str <python:str>` or :obj:`None <python:None>`
        """
        return self._height

    @height.setter
    def height(self, value):
        self._height = validators.string(value, allow_empty = True)

    @property
    def high_index(self) -> Optional[str]:
        """Defaults to ``'High index'``.

        :rtype: :class:`str <python:str>` or :obj:`None <python:None>`
        """
        return self._high_index

    @high_index.setter
    def high_index(self, value):
        self._high_index = validators.string(value, allow_empty = True)

    @property
    def horizontal_line(self) -> Optional[str]:
        """Defaults to ``'Horizontal line'``.

        :rtype: :class:`str <python:str>` or :obj:`None <python:None>`
        """
        return self._horizontal_line

    @horizontal_line.setter
    def horizontal_line(self, value):
        self._horizontal_line = validators.string(value, allow_empty = True)

    @property
    def increment(self) -> Optional[str]:
        """Defaults to ``'Increment``.

        :rtype: :class:`str <python:str>` or :obj:`None <python:None>`
        """
        return self._increment

    @increment.setter
    def increment(self, value):
        self._increment = validators.string(value, allow_empty = True)

    @property
    def index(self) -> Optional[str]:
        """Defaults to ``'Index``.

        :rtype: :class:`str <python:str>` or :obj:`None <python:None>`
        """
        return self._index

    @index.setter
    def index(self, value):
        self._index = validators.string(value, allow_empty = True)

    @property
    def infinity_line(self) -> Optional[str]:
        """Defaults to ``'Infinity line'``.

        :rtype: :class:`str <python:str>` or :obj:`None <python:None>`
        """
        return self._infinity_line

    @infinity_line.setter
    def infinity_line(self, value):
        self._infinity_line = validators.string(value, allow_empty = True)

    @property
    def initial_acceleration_factor(self) -> Optional[str]:
        """Defaults to
        ``'Initial acceleration factor'``.

        :rtype: :class:`str <python:str>` or :obj:`None <python:None>`
        """
        return self._initial_acceleration_factor

    @initial_acceleration_factor.setter
    def initial_acceleration_factor(self, value):
        self._initial_acceleration_factor = validators.string(value, allow_empty = True)

    @property
    def inner_background(self) -> Optional[str]:
        """Defaults to ``'Inner background'``.

        :rtype: :class:`str <python:str>` or :obj:`None <python:None>`
        """
        return self._inner_background

    @inner_background.setter
    def inner_background(self, value):
        self._inner_background = validators.string(value, allow_empty = True)

    @property
    def label(self) -> Optional[str]:
        """Defaults to ``'Label``.

        :rtype: :class:`str <python:str>` or :obj:`None <python:None>`
        """
        return self._label

    @label.setter
    def label(self, value):
        self._label = validators.string(value, allow_empty = True)

    @property
    def label_options(self) -> Optional[str]:
        """Defaults to ``'Label options'``.

        :rtype: :class:`str <python:str>` or :obj:`None <python:None>`
        """
        return self._label_options

    @label_options.setter
    def label_options(self, value):
        self._label_options = validators.string(value, allow_empty = True)

    @property
    def labels(self) -> Optional[str]:
        """Defaults to ``'Labels``.

        :rtype: :class:`str <python:str>` or :obj:`None <python:None>`
        """
        return self._labels

    @labels.setter
    def labels(self, value):
        self._labels = validators.string(value, allow_empty = True)

    @property
    def line(self) -> Optional[str]:
        """Defaults to ``'Line``.

        :rtype: :class:`str <python:str>` or :obj:`None <python:None>`
        """
        return self._line

    @line.setter
    def line(self, value):
        self._line = validators.string(value, allow_empty = True)

    @property
    def lines(self) -> Optional[str]:
        """Defaults to ``'Lines``.

        :rtype: :class:`str <python:str>` or :obj:`None <python:None>`
        """
        return self._lines

    @lines.setter
    def lines(self, value):
        self._lines = validators.string(value, allow_empty = True)

    @property
    def long_period(self) -> Optional[str]:
        """Defaults to ``'Long period'``.

        :rtype: :class:`str <python:str>` or :obj:`None <python:None>`
        """
        return self._long_period

    @long_period.setter
    def long_period(self, value):
        self._long_period = validators.string(value, allow_empty = True)

    @property
    def low_index(self) -> Optional[str]:
        """Defaults to ``'Low index'``.

        :rtype: :class:`str <python:str>` or :obj:`None <python:None>`
        """
        return self._low_index

    @low_index.setter
    def low_index(self, value):
        self._low_index = validators.string(value, allow_empty = True)

    @property
    def max_acceleration_factor(self) -> Optional[str]:
        """Defaults to
        ``'Max acceleration factor'``.

        :rtype: :class:`str <python:str>` or :obj:`None <python:None>`
        """
        return self._max_acceleration_factor

    @max_acceleration_factor.setter
    def max_acceleration_factor(self, value):
        self._max_acceleration_factor = validators.string(value, allow_empty = True)

    @property
    def measure(self) -> Optional[str]:
        """Defaults to ``'Measure``.

        :rtype: :class:`str <python:str>` or :obj:`None <python:None>`
        """
        return self._measure

    @measure.setter
    def measure(self, value):
        self._measure = validators.string(value, allow_empty = True)

    @property
    def measure_x(self) -> Optional[str]:
        """Defaults to ``'Measure X'``.

        :rtype: :class:`str <python:str>` or :obj:`None <python:None>`
        """
        return self._measure_x

    @measure_x.setter
    def measure_x(self, value):
        self._measure_x = validators.string(value, allow_empty = True)

    @property
    def measure_xy(self) -> Optional[str]:
        """Defaults to ``'Measure XY'``.

        :rtype: :class:`str <python:str>` or :obj:`None <python:None>`
        """
        return self._measure_xy

    @measure_xy.setter
    def measure_xy(self, value):
        self._measure_xy = validators.string(value, allow_empty = True)

    @property
    def measure_y(self) -> Optional[str]:
        """Defaults to ``'Measure Y'``.

        :rtype: :class:`str <python:str>` or :obj:`None <python:None>`
        """
        return self._measure_y

    @measure_y.setter
    def measure_y(self, value):
        self._measure_y = validators.string(value, allow_empty = True)

    @property
    def multiplier(self) -> Optional[str]:
        """Defaults to ``'Multiplier``.

        :rtype: :class:`str <python:str>` or :obj:`None <python:None>`
        """
        return self._multiplier

    @multiplier.setter
    def multiplier(self, value):
        self._multiplier = validators.string(value, allow_empty = True)

    @property
    def multiplier_atr(self) -> Optional[str]:
        """Defaults to ``'ATR multiplier'``.

        :rtype: :class:`str <python:str>` or :obj:`None <python:None>`
        """
        return self._multiplier_atr

    @multiplier_atr.setter
    def multiplier_atr(self, value):
        self._multiplier_atr = validators.string(value, allow_empty = True)

    @property
    def name(self) -> Optional[str]:
        """Defaults to ``'Name``.

        :rtype: :class:`str <python:str>` or :obj:`None <python:None>`
        """
        return self._name

    @name.setter
    def name(self, value):
        self._name = validators.string(value, allow_empty = True)

    @property
    def no_filter_match(self) -> Optional[str]:
        """Defaults to ``'No match'``.

        :rtype: :class:`str <python:str>` or :obj:`None <python:None>`
        """
        return self._no_filter_match

    @no_filter_match.setter
    def no_filter_match(self, value):
        self._no_filter_match = validators.string(value, allow_empty = True)

    @property
    def outer_background(self) -> Optional[str]:
        """Defaults to ``'Outer background'``.

        :rtype: :class:`str <python:str>` or :obj:`None <python:None>`
        """
        return self._outer_background

    @outer_background.setter
    def outer_background(self, value):
        self._outer_background = validators.string(value, allow_empty = True)

    @property
    def padding(self) -> Optional[str]:
        """Defaults to ``'Padding``.

        :rtype: :class:`str <python:str>` or :obj:`None <python:None>`
        """
        return self._padding

    @padding.setter
    def padding(self, value):
        self._padding = validators.string(value, allow_empty = True)

    @property
    def parallel_channel(self) -> Optional[str]:
        """Defaults to ``'Parallel channel'``.

        :rtype: :class:`str <python:str>` or :obj:`None <python:None>`
        """
        return self._parallel_channel

    @parallel_channel.setter
    def parallel_channel(self, value):
        self._parallel_channel = validators.string(value, allow_empty = True)

    @property
    def period(self) -> Optional[str]:
        """Defaults to ``'Period``.

        :rtype: :class:`str <python:str>` or :obj:`None <python:None>`
        """
        return self._period

    @period.setter
    def period(self, value):
        self._period = validators.string(value, allow_empty = True)

    @property
    def period_atr(self) -> Optional[str]:
        """Defaults to ``'ATR period'``.

        :rtype: :class:`str <python:str>` or :obj:`None <python:None>`
        """
        return self._period_atr

    @period_atr.setter
    def period_atr(self, value):
        self._period_atr = validators.string(value, allow_empty = True)

    @property
    def periods(self) -> Optional[str]:
        """Defaults to ``'Periods``.

        :rtype: :class:`str <python:str>` or :obj:`None <python:None>`
        """
        return self._periods

    @periods.setter
    def periods(self, value):
        self._periods = validators.string(value, allow_empty = True)

    @property
    def period_senkou_span_b(self) -> Optional[str]:
        """Defaults to
        ``'Senkou Span B period'``.

        :rtype: :class:`str <python:str>` or :obj:`None <python:None>`
        """
        return self._period_senkou_span_b

    @period_senkou_span_b.setter
    def period_senkou_span_b(self, value):
        self._period_senkou_span_b = validators.string(value, allow_empty = True)

    @property
    def period_tenkan(self) -> Optional[str]:
        """Defaults to ``'Tenkan period'``.

        :rtype: :class:`str <python:str>` or :obj:`None <python:None>`
        """
        return self._period_tenkan

    @period_tenkan.setter
    def period_tenkan(self, value):
        self._period_tenkan = validators.string(value, allow_empty = True)

    @property
    def pitchfork(self) -> Optional[str]:
        """Defaults to ``'Pitchfork``.

        :rtype: :class:`str <python:str>` or :obj:`None <python:None>`
        """
        return self._pitchfork

    @pitchfork.setter
    def pitchfork(self, value):
        self._pitchfork = validators.string(value, allow_empty = True)

    @property
    def ranges(self) -> Optional[str]:
        """Defaults to ``'Ranges``.

        :rtype: :class:`str <python:str>` or :obj:`None <python:None>`
        """
        return self._ranges

    @ranges.setter
    def ranges(self, value):
        self._ranges = validators.string(value, allow_empty = True)

    @property
    def ray(self) -> Optional[str]:
        """Defaults to ``'Ray``.

        :rtype: :class:`str <python:str>` or :obj:`None <python:None>`
        """
        return self._ray

    @ray.setter
    def ray(self, value):
        self._ray = validators.string(value, allow_empty = True)

    @property
    def rectangle(self) -> Optional[str]:
        """Defaults to ``'Rectangle``.

        :rtype: :class:`str <python:str>` or :obj:`None <python:None>`
        """
        return self._rectangle

    @rectangle.setter
    def rectangle(self, value):
        self._rectangle = validators.string(value, allow_empty = True)

    @property
    def remove_button(self) -> Optional[str]:
        """Defaults to ``'remove'``.

        :rtype: :class:`str <python:str>` or :obj:`None <python:None>`
        """
        return self._remove_button

    @remove_button.setter
    def remove_button(self, value):
        self._remove_button = validators.string(value, allow_empty = True)

    @property
    def save_button(self) -> Optional[str]:
        """Defaults to ``'save'``.

        :rtype: :class:`str <python:str>` or :obj:`None <python:None>`
        """
        return self._save_button

    @save_button.setter
    def save_button(self, value):
        self._save_button = validators.string(value, allow_empty = True)

    @property
    def search_indicators(self) -> Optional[str]:
        """Defaults to
        ``'Search Indicators'``.

        :rtype: :class:`str <python:str>` or :obj:`None <python:None>`
        """
        return self._search_indicators

    @search_indicators.setter
    def search_indicators(self, value):
        self._search_indicators = validators.string(value, allow_empty = True)

    @property
    def segment(self) -> Optional[str]:
        """Defaults to ``'Segment``.

        :rtype: :class:`str <python:str>` or :obj:`None <python:None>`
        """
        return self._segment

    @segment.setter
    def segment(self, value):
        self._segment = validators.string(value, allow_empty = True)

    @property
    def series(self) -> Optional[str]:
        """Defaults to ``'Series``.

        :rtype: :class:`str <python:str>` or :obj:`None <python:None>`
        """
        return self._series

    @series.setter
    def series(self, value):
        self._series = validators.string(value, allow_empty = True)

    @property
    def shape_options(self) -> Optional[str]:
        """Defaults to ``'Shape options'``.

        :rtype: :class:`str <python:str>` or :obj:`None <python:None>`
        """
        return self._shape_options

    @shape_options.setter
    def shape_options(self, value):
        self._shape_options = validators.string(value, allow_empty = True)

    @property
    def shapes(self) -> Optional[str]:
        """Defaults to ``'Shapes``.

        :rtype: :class:`str <python:str>` or :obj:`None <python:None>`
        """
        return self._shapes

    @shapes.setter
    def shapes(self, value):
        self._shapes = validators.string(value, allow_empty = True)

    @property
    def short_period(self) -> Optional[str]:
        """Defaults to ``'Short period'``.

        :rtype: :class:`str <python:str>` or :obj:`None <python:None>`
        """
        return self._short_period

    @short_period.setter
    def short_period(self, value):
        self._short_period = validators.string(value, allow_empty = True)

    @property
    def signal_period(self) -> Optional[str]:
        """Defaults to ``'Signal period'``.

        :rtype: :class:`str <python:str>` or :obj:`None <python:None>`
        """
        return self._signal_period

    @signal_period.setter
    def signal_period(self, value):
        self._signal_period = validators.string(value, allow_empty = True)

    @property
    def simple_shapes(self) -> Optional[str]:
        """Defaults to ``'Simple shapes'``.

        :rtype: :class:`str <python:str>` or :obj:`None <python:None>`
        """
        return self._simple_shapes

    @simple_shapes.setter
    def simple_shapes(self, value):
        self._simple_shapes = validators.string(value, allow_empty = True)

    @property
    def slow_avg_period(self) -> Optional[str]:
        """Defaults to ``'Slow average period'``.

        :rtype: :class:`str <python:str>` or :obj:`None <python:None>`
        """
        return self._slow_avg_period

    @slow_avg_period.setter
    def slow_avg_period(self, value):
        self._slow_avg_period = validators.string(value, allow_empty = True)

    @property
    def standard_deviation(self) -> Optional[str]:
        """Defaults to
        ``'Standard deviation'``.

        :rtype: :class:`str <python:str>` or :obj:`None <python:None>`
        """
        return self._standard_deviation

    @standard_deviation.setter
    def standard_deviation(self, value):
        self._standard_deviation = validators.string(value, allow_empty = True)

    @property
    def stroke(self) -> Optional[str]:
        """Defaults to ``'Line color'``.

        :rtype: :class:`str <python:str>` or :obj:`None <python:None>`
        """
        return self._stroke

    @stroke.setter
    def stroke(self, value):
        self._stroke = validators.string(value, allow_empty = True)

    @property
    def stroke_width(self) -> Optional[str]:
        """Defaults to ``'Line width'``.

        :rtype: :class:`str <python:str>` or :obj:`None <python:None>`
        """
        return self._stroke_width

    @stroke_width.setter
    def stroke_width(self, value):
        self._stroke_width = validators.string(value, allow_empty = True)

    @property
    def style(self) -> Optional[str]:
        """Defaults to ``'Style``.

        :rtype: :class:`str <python:str>` or :obj:`None <python:None>`
        """
        return self._style

    @style.setter
    def style(self, value):
        self._style = validators.string(value, allow_empty = True, coerce_value = True)

    @property
    def time_cycles(self) -> Optional[str]:
        """Defaults to ``'Time Cycles'``.

        :rtype: :class:`str <python:str>` or :obj:`None <python:None>`
        """
        return self._time_cycles

    @time_cycles.setter
    def time_cycles(self, value):
        self._time_cycles = validators.string(value, allow_empty = True)

    @property
    def title(self) -> Optional[str]:
        """Defaults to ``'Title``.

        :rtype: :class:`str <python:str>` or :obj:`None <python:None>`
        """
        return self._title

    @title.setter
    def title(self, value):
        self._title = validators.string(value, allow_empty = True)

    @property
    def top_band(self) -> Optional[str]:
        """Defaults to ``'Top band'``.

        :rtype: :class:`str <python:str>` or :obj:`None <python:None>`
        """
        return self._top_band

    @top_band.setter
    def top_band(self, value):
        self._top_band = validators.string(value, allow_empty = True)

    @property
    def tunnel(self) -> Optional[str]:
        """Defaults to ``'Tunnel``.

        :rtype: :class:`str <python:str>` or :obj:`None <python:None>`
        """
        return self._tunnel

    @tunnel.setter
    def tunnel(self, value):
        self._tunnel = validators.string(value, allow_empty = True)

    @property
    def type_options(self) -> Optional[str]:
        """Defaults to ``'Details'``.

        :rtype: :class:`str <python:str>` or :obj:`None <python:None>`
        """
        return self._type_options

    @type_options.setter
    def type_options(self, value):
        self._type_options = validators.string(value, allow_empty = True)

    @property
    def vertical_arrow(self) -> Optional[str]:
        """Defaults to ``'Vertical arrow'``.

        :rtype: :class:`str <python:str>` or :obj:`None <python:None>`
        """
        return self._vertical_arrow

    @vertical_arrow.setter
    def vertical_arrow(self, value):
        self._vertical_arrow = validators.string(value, allow_empty = True)

    @property
    def vertical_counter(self) -> Optional[str]:
        """Defaults to ``'Vertical counter'``.

        :rtype: :class:`str <python:str>` or :obj:`None <python:None>`
        """
        return self._vertical_counter

    @vertical_counter.setter
    def vertical_counter(self, value):
        self._vertical_counter = validators.string(value, allow_empty = True)

    @property
    def vertical_label(self) -> Optional[str]:
        """Defaults to ``'Vertical label'``.

        :rtype: :class:`str <python:str>` or :obj:`None <python:None>`
        """
        return self._vertical_label

    @vertical_label.setter
    def vertical_label(self, value):
        self._vertical_label = validators.string(value, allow_empty = True)

    @property
    def vertical_line(self) -> Optional[str]:
        """Defaults to ``'Vertical line'``.

        :rtype: :class:`str <python:str>` or :obj:`None <python:None>`
        """
        return self._vertical_line

    @vertical_line.setter
    def vertical_line(self, value):
        self._vertical_line = validators.string(value, allow_empty = True)

    @property
    def volume(self) -> Optional[str]:
        """Defaults to ``'Volume'``.

        :rtype: :class:`str <python:str>` or :obj:`None <python:None>`
        """
        return self._volume

    @volume.setter
    def volume(self, value):
        self._volume = validators.string(value, allow_empty = True)

    @property
    def x_axis_unit(self) -> Optional[str]:
        """Defaults to ``'x-axis unit'``.

        :rtype: :class:`str <python:str>` or :obj:`None <python:None>`
        """
        return self._x_axis_unit

    @x_axis_unit.setter
    def x_axis_unit(self, value):
        self._x_axis_unit = validators.string(value, allow_empty = True)

    @classmethod
    def _get_kwargs_from_dict(cls, as_dict):
        kwargs = {
            'add_button': as_dict.get('addButton', None),
            'algorithm': as_dict.get('algorithm', None),
            'arrow_infinity_line': as_dict.get('arrowInfinityLine', None),
            'arrow_ray': as_dict.get('arrowRay', None),
            'arrow_segment': as_dict.get('arrowSegment', None),
            'average': as_dict.get('average', None),
            'background': as_dict.get('background', None),
            'background_color': as_dict.get('backgroundColor', None),
            'background_colors': as_dict.get('backgroundColors', None),
            'border_color': as_dict.get('borderColor', None),
            'border_radius': as_dict.get('borderRadius', None),
            'border_width': as_dict.get('borderWidth', None),
            'bottom_band': as_dict.get('bottomBand', None),
            'circle': as_dict.get('circle', None),
            'clear_filter': as_dict.get('clearFilter', None),
            'color': as_dict.get('color', None),
            'connector': as_dict.get('connector', None),
            'crooked3': as_dict.get('crooked3', None),
            'crooked5': as_dict.get('crooked5', None),
            'crosshairX': as_dict.get('crosshairX', None),
            'crosshairY': as_dict.get('crosshairY', None),
            'decimals': as_dict.get('decimals', None),
            'deviation': as_dict.get('deviation', None),
            'edit_button': as_dict.get('editButton', None),
            'elliott3': as_dict.get('elliott3', None),
            'elliott5': as_dict.get('elliott5', None),
            'ellipse': as_dict.get('ellipse', None),
            'factor': as_dict.get('factor', None),
            'fast_avg_period': as_dict.get('fastAvgPeriod', None),
            'fibonacci': as_dict.get('fibonacci', None),
            'fibonacci_time_zones': as_dict.get('fibonacciTimeZones', None),
            'fill': as_dict.get('fill', None),
            'flags': as_dict.get('flags', None),
            'font_size': as_dict.get('fontSize', None),
            'format': as_dict.get('format', None),
            'height': as_dict.get('height', None),
            'high_index': as_dict.get('highIndex', None),
            'horizontal_line': as_dict.get('horizontalLine', None),
            'increment': as_dict.get('increment', None),
            'index': as_dict.get('index', None),
            'infinity_line': as_dict.get('infinityLine', None),
            'initial_acceleration_factor': as_dict.get('initialAccelerationFactor', None),
            'inner_background': as_dict.get('innerBackground', None),
            'label': as_dict.get('label', None),
            'label_options': as_dict.get('labelOptions', None),
            'labels': as_dict.get('labels', None),
            'line': as_dict.get('line', None),
            'lines': as_dict.get('lines', None),
            'long_period': as_dict.get('longPeriod', None),
            'low_index': as_dict.get('lowIndex', None),
            'max_acceleration_factor': as_dict.get('maxAccelerationFactor', None),
            'measure': as_dict.get('measure', None),
            'measure_x': as_dict.get('measureX', None),
            'measure_xy': as_dict.get('measureXY', None),
            'measure_y': as_dict.get('measureY', None),
            'multiplier': as_dict.get('multiplier', None),
            'multiplier_atr': as_dict.get('multiplierATR', None),
            'name': as_dict.get('name', None),
            'no_filter_match': as_dict.get('noFilterMatch', None),
            'outer_background': as_dict.get('outerBackground', None),
            'padding': as_dict.get('padding', None),
            'parallel_channel': as_dict.get('parallelChannel', None),
            'period': as_dict.get('period', None),
            'period_atr': as_dict.get('periodATR', None),
            'periods': as_dict.get('periods', None),
            'period_senkou_span_b': as_dict.get('periodSenkouSpanB', None),
            'period_tenkan': as_dict.get('periodTenkan', None),
            'pitchfork': as_dict.get('pitchfork', None),
            'ranges': as_dict.get('ranges', None),
            'ray': as_dict.get('ray', None),
            'rectangle': as_dict.get('rectangle', None),
            'remove_button': as_dict.get('removeButton', None),
            'save_button': as_dict.get('saveButton', None),
            'search_indicators': as_dict.get('searchIndicators', None),
            'segment': as_dict.get('segment', None),
            'series': as_dict.get('series', None),
            'shape_options': as_dict.get('shapeOptions', None),
            'shapes': as_dict.get('shapes', None),
            'short_period': as_dict.get('shortPeriod', None),
            'signal_period': as_dict.get('signalPeriod', None),
            'simple_shapes': as_dict.get('simpleShapes', None),
            'slow_avg_period': as_dict.get('slowAvgPeriod', None),
            'standard_deviation': as_dict.get('standardDeviation', None),
            'stroke': as_dict.get('stroke', None),
            'stroke_width': as_dict.get('strokeWidth', None),
            'style': as_dict.get('style', None),
            'time_cycles': as_dict.get('timeCycles', None),
            'title': as_dict.get('title', None),
            'top_band': as_dict.get('topBand', None),
            'tunnel': as_dict.get('tunnel', None),
            'type_options': as_dict.get('typeOptions', None),
            'vertical_arrow': as_dict.get('verticalArrow', None),
            'vertical_counter': as_dict.get('verticalCounter', None),
            'vertical_label': as_dict.get('verticalLabel', None),
            'vertical_line': as_dict.get('verticalLine', None),
            'volume': as_dict.get('volume', None),
            'x_axis_unit': as_dict.get('xAxisUnit', None),
        }

        return kwargs

    def _to_untrimmed_dict(self, in_cls = None) -> dict:
        untrimmed = {
            'addButton': self.add_button,
            'algorithm': self.algorithm,
            'arrowInfinityLine': self.arrow_infinity_line,
            'arrowRay': self.arrow_ray,
            'arrowSegment': self.arrow_segment,
            'average': self.average,
            'background': self.background,
            'backgroundColor': self.background_color,
            'backgroundColors': self.background_colors,
            'borderColor': self.border_color,
            'borderRadius': self.border_radius,
            'borderWidth': self.border_width,
            'bottomBand': self.bottom_band,
            'circle': self.circle,
            'clearFilter': self.clear_filter,
            'color': self.color,
            'connector': self.connector,
            'crooked3': self.crooked3,
            'crooked5': self.crooked5,
            'crosshairX': self.crosshairX,
            'crosshairY': self.crosshairY,
            'decimals': self.decimals,
            'deviation': self.deviation,
            'editButton': self.edit_button,
            'elliott3': self.elliott3,
            'elliott5': self.elliott5,
            'ellipse': self.ellipse,
            'factor': self.factor,
            'fastAvgPeriod': self.fast_avg_period,
            'fibonacci': self.fibonacci,
            'fibonacciTimeZones': self.fibonacci_time_zones,
            'fill': self.fill,
            'flags': self.flags,
            'fontSize': self.font_size,
            'format': self.format,
            'height': self.height,
            'highIndex': self.high_index,
            'horizontalLine': self.horizontal_line,
            'increment': self.increment,
            'index': self.index,
            'infinityLine': self.infinity_line,
            'initialAccelerationFactor': self.initial_acceleration_factor,
            'innerBackground': self.inner_background,
            'label': self.label,
            'labelOptions': self.label_options,
            'labels': self.labels,
            'line': self.line,
            'lines': self.lines,
            'longPeriod': self.long_period,
            'lowIndex': self.low_index,
            'maxAccelerationFactor': self.max_acceleration_factor,
            'measure': self.measure,
            'measureX': self.measure_x,
            'measureXY': self.measure_xy,
            'measureY': self.measure_y,
            'multiplier': self.multiplier,
            'multiplierATR': self.multiplier_atr,
            'name': self.name,
            'noFilterMatch': self.no_filter_match,
            'outerBackground': self.outer_background,
            'padding': self.padding,
            'parallelChannel': self.parallel_channel,
            'period': self.period,
            'periodATR': self.period_atr,
            'periods': self.periods,
            'periodSenkouSpanB': self.period_senkou_span_b,
            'periodTenkan': self.period_tenkan,
            'pitchfork': self.pitchfork,
            'ranges': self.ranges,
            'ray': self.ray,
            'rectangle': self.rectangle,
            'removeButton': self.remove_button,
            'saveButton': self.save_button,
            'searchIndicators': self.search_indicators,
            'segment': self.segment,
            'series': self.series,
            'shapeOptions': self.shape_options,
            'shapes': self.shapes,
            'shortPeriod': self.short_period,
            'signalPeriod': self.signal_period,
            'simpleShapes': self.simple_shapes,
            'slowAvgPeriod': self.slow_avg_period,
            'standardDeviation': self.standard_deviation,
            'stroke': self.stroke,
            'strokeWidth': self.stroke_width,
            'style': self.style,
            'timeCycles': self.time_cycles,
            'title': self.title,
            'topBand': self.top_band,
            'tunnel': self.tunnel,
            'typeOptions': self.type_options,
            'verticalArrow': self.vertical_arrow,
            'verticalCounter': self.vertical_counter,
            'verticalLabel': self.vertical_label,
            'verticalLine': self.vertical_line,
            'volume': self.volume,
            'xAxisUnit': self.x_axis_unit
        }

        return untrimmed


class NavigationLanguageOptions(HighchartsMeta):
    """The Popup strings used in the chart.

    .. note::

      Requires the ``annotations.js`` or ``annotations-advanced.src.js`` module to be
      loaded.

    """

    def __init__(self, **kwargs):
        self._popup = None

        self.popup = kwargs.get('popup', None)

    @property
    def popup(self) -> Optional[PopupLanguageOptions]:
        """Translations for all field names used in popup.

        :rtype: :class:`PopupLanguageOptions` or :obj:`None <python:None>`
        """
        return self._popup

    @popup.setter
    @class_sensitive(PopupLanguageOptions)
    def popup(self, value):
        self._popup = value

    @classmethod
    def _get_kwargs_from_dict(cls, as_dict):
        kwargs = {
            'popup': as_dict.get('popup', None)
        }

        return kwargs

    def _to_untrimmed_dict(self, in_cls = None) -> dict:
        return {
            'popup': self.popup
        }



from typing import Optional
from decimal import Decimal

from validator_collection import validators

from highcharts_core import constants, errors
from highcharts_core.decorators import class_sensitive
from highcharts_core.metaclasses import HighchartsMeta
from highcharts_core.utility_classes.gradients import Gradient
from highcharts_core.utility_classes.patterns import Pattern
from highcharts_core.utility_classes.javascript_functions import CallbackFunction


class DragHandle(HighchartsMeta):
    """Options for the drag handles available in column series."""

    def __init__(self, **kwargs):
        self._class_name = None
        self._color = None
        self._cursor = None
        self._line_color = None
        self._line_width = None
        self._path_formatter = None
        self._z_index = None

        self.class_name = kwargs.get('class_name', None)
        self.color = kwargs.get('color', None)
        self.cursor = kwargs.get('cursor', None)
        self.line_color = kwargs.get('line_color', None)
        self.line_width = kwargs.get('line_width', None)
        self.path_formatter = kwargs.get('path_formatter', None)
        self.z_index = kwargs.get('z_index', None)

    @property
    def class_name(self) -> Optional[str]:
        """CSS class name of the guide box in this state. Defaults to
        ``'highcharts-drag-handle'``.

        :rtype: :class:`str <python:str>` or :obj:`None <python:None>`
        """
        return self._class_name

    @class_name.setter
    def class_name(self, value):
        self._class_name = validators.string(value, allow_empty = True)

    @property
    def color(self) -> Optional[str | Gradient | Pattern]:
        """The fill color of the drag handles. Defaults to ``'#fff'``.

        :rtype: :class:`str <python:str>`, :class:`Gradient`, :class:`Pattern``, or
          :obj:`None <python:None>`
        """
        return self._color

    @color.setter
    def color(self, value):
        from highcharts_core import utility_functions
        self._color = utility_functions.validate_color(value)

    @property
    def cursor(self) -> Optional[str]:
        """The mouse cursor to use for the drag handles. By default (when
        :obj:`None <python:None>`), this intelligently switches between ``'ew-resize'``
        and ``'ns-resize'`` depending on the direction the point is being dragged.

        Acceptable values are:

          * ``'alias'``
          * ``'all-scroll'``
          * ``'auto'``
          * ``'cell'``
          * ``'col-resize'``
          * ``'context-menu'``
          * ``'copy'``
          * ``'crosshair'``
          * ``'default'``
          * ``'e-resize'``
          * ``'ew-resize'``
          * ``'grab'``
          * ``'grabbing'``
          * ``'help'``
          * ``'move'``
          * ``'n-resize'``
          * ``'ne-resize'``
          * ``'nesw-resize'``
          * ``'no-drop'``
          * ``'none'``
          * ``'not-allowed'``
          * ``'ns-resize'``
          * ``'nw-resize'``
          * ``'nwse-resize'``
          * ``'pointer'``
          * ``'progress'``
          * ``'row-resize'``
          * ``'s-resize'``
          * ``'se-resize'``
          * ``'sw-resize'``
          * ``'text'``
          * ``'vertical-text'``
          * ``'w-resize'``
          * ``'wait'``
          * ``'zoom-in'``
          * ``'zoom-out'``

        :rtype: :class:`str <python:str>` or :obj:`None <python:None>`
        """
        return self._cursor

    @cursor.setter
    def cursor(self, value):
        if not value:
            self._cursor = None
        else:
            value = validators.string(value)
            value = value.lower()
            if value not in constants.SUPPORTED_CURSOR_VALUES:
                raise errors.HighchartsValueError(f'cursor expects a valid cursor value. '
                                                  f'Received: {value}')
            self._cursor = value

    @property
    def line_color(self) -> Optional[str | Gradient | Pattern]:
        """The line color of the drag handles. Defaults to ``'rgba(0, 0, 0, 0.6)'``.

        :rtype: :obj:`None <python:None>`, :class:`Gradient`, :class:`Pattern`, or
          :class:`str <python:str>`
        """
        return self._line_color

    @line_color.setter
    def line_color(self, value):
        from highcharts_core import utility_functions
        self._line_color = utility_functions.validate_color(value)

    @property
    def line_width(self) -> Optional[int | float | Decimal]:
        """The line width for the drag handles. Defaults to ``1``.

        :rtype: numeric or :obj:`None <python:None>`
        """
        return self._line_width

    @line_width.setter
    def line_width(self, value):
        self._line_width = validators.numeric(value,
                                              allow_empty = True,
                                              minimum = 0)

    @property
    def path_formatter(self) -> Optional[CallbackFunction]:
        """JavaScript function to define the SVG path to use for the drag handles. Takes
        the ``point`` as a (JavaScript) argument. Should return an SVG path in array
        format. The SVG path is automatically positioned on the point.

        :rtype: :class:`CallbackFunction` or :obj:`None <python:None>`
        """
        return self._path_formatter

    @path_formatter.setter
    @class_sensitive(CallbackFunction)
    def path_formatter(self, value):
        self._path_formatter = value

    @property
    def z_index(self) -> Optional[int | float | Decimal]:
        """Drag handles' zIndex position. Defaults to ``901``.

        :rtype: numeric or :obj:`None <python:None>`
        """
        return self._z_index

    @z_index.setter
    def z_index(self, value):
        self._z_index = validators.numeric(value, allow_empty = True)

    @classmethod
    def _get_kwargs_from_dict(cls, as_dict):
        kwargs = {
            'class_name': as_dict.get('className', None),
            'color': as_dict.get('color', None),
            'cursor': as_dict.get('cursor', None),
            'line_color': as_dict.get('lineColor', None),
            'line_width': as_dict.get('lineWidth', None),
            'path_formatter': as_dict.get('pathFormatter', None),
            'z_index': as_dict.get('zIndex', None),
        }

        return kwargs

    def _to_untrimmed_dict(self, in_cls = None) -> dict:
        untrimmed = {
            'className': self.class_name,
            'color': self.color,
            'cursor': self.cursor,
            'lineColor': self.line_color,
            'lineWidth': self.line_width,
            'pathFormatter': self.path_formatter,
            'zIndex': self.z_index
        }

        return untrimmed


class GuideBoxOptions(HighchartsMeta):
    """Style options for the guide box default state."""

    def __init__(self, **kwargs):
        self._class_name = None
        self._color = None
        self._cursor = None
        self._line_color = None
        self._line_width = None
        self._z_index = None

        self.class_name = kwargs.get('class_name', None)
        self.color = kwargs.get('color', None)
        self.cursor = kwargs.get('cursor', None)
        self.line_color = kwargs.get('line_color', None)
        self.line_width = kwargs.get('line_width', None)
        self.z_index = kwargs.get('z_index', None)

    @property
    def class_name(self) -> Optional[str]:
        """CSS class name of the guide box in this state. Defaults to
        ``'highcharts-drag-box-default'``.

        :rtype: :class:`str <python:str>` or :obj:`None <python:None>`
        """
        return self._class_name

    @class_name.setter
    def class_name(self, value):
        self._class_name = validators.string(value, allow_empty = True)

    @property
    def color(self) -> Optional[str | Gradient | Pattern]:
        """Guide box fill color. Defaults to ``'rgba(0, 0, 0, 0.1)'``.

        :rtype: :class:`str <python:str>`, :class:`Gradient`, :class:`Pattern``, or
          :obj:`None <python:None>`
        """
        return self._color

    @color.setter
    def color(self, value):
        from highcharts_core import utility_functions
        self._color = utility_functions.validate_color(value)

    @property
    def cursor(self) -> Optional[str]:
        """The style of cursor to use when the user's mouse hovers over the data series.
        Defaults to ``'move'``.

        Acceptable values are:

          * ``'alias'``
          * ``'all-scroll'``
          * ``'auto'``
          * ``'cell'``
          * ``'col-resize'``
          * ``'context-menu'``
          * ``'copy'``
          * ``'crosshair'``
          * ``'default'``
          * ``'e-resize'``
          * ``'ew-resize'``
          * ``'grab'``
          * ``'grabbing'``
          * ``'help'``
          * ``'move'``
          * ``'n-resize'``
          * ``'ne-resize'``
          * ``'nesw-resize'``
          * ``'no-drop'``
          * ``'none'``
          * ``'not-allowed'``
          * ``'ns-resize'``
          * ``'nw-resize'``
          * ``'nwse-resize'``
          * ``'pointer'``
          * ``'progress'``
          * ``'row-resize'``
          * ``'s-resize'``
          * ``'se-resize'``
          * ``'sw-resize'``
          * ``'text'``
          * ``'vertical-text'``
          * ``'w-resize'``
          * ``'wait'``
          * ``'zoom-in'``
          * ``'zoom-out'``

        :rtype: :class:`str <python:str>` or :obj:`None <python:None>`
        """
        return self._cursor

    @cursor.setter
    def cursor(self, value):
        if not value:
            self._cursor = None
        else:
            value = validators.string(value)
            value = value.lower()
            if value not in constants.SUPPORTED_CURSOR_VALUES:
                raise errors.HighchartsValueError(f'cursor expects a valid cursor value. '
                                                  f'Received: {value}')
            self._cursor = value

    @property
    def line_color(self) -> Optional[str | Gradient | Pattern]:
        """Color of the border around the guide box. Defaults to ``'#888'``.

        :rtype: :obj:`None <python:None>`, :class:`Gradient`, :class:`Pattern`, or
          :class:`str <python:str>`
        """
        return self._line_color

    @line_color.setter
    def line_color(self, value):
        from highcharts_core import utility_functions
        self._line_color = utility_functions.validate_color(value)

    @property
    def line_width(self) -> Optional[int | float | Decimal]:
        """Pixel width of the line around the guide box. Defaults to ``1``.

        :rtype: numeric or :obj:`None <python:None>`
        """
        return self._line_width

    @line_width.setter
    def line_width(self, value):
        self._line_width = validators.numeric(value,
                                              allow_empty = True,
                                              minimum = 0)

    @property
    def z_index(self) -> Optional[int | float | Decimal]:
        """Guide box zIndex position. Defaults to ``900``.

        :rtype: numeric or :obj:`None <python:None>`
        """
        return self._z_index

    @z_index.setter
    def z_index(self, value):
        self._z_index = validators.numeric(value, allow_empty = True)

    @classmethod
    def _get_kwargs_from_dict(cls, as_dict):
        kwargs = {
            'class_name': as_dict.get('className', None),
            'color': as_dict.get('color', None),
            'cursor': as_dict.get('cursor', None),
            'line_color': as_dict.get('lineColor', None),
            'line_width': as_dict.get('lineWidth', None),
            'z_index': as_dict.get('zIndex', None),
        }

        return kwargs

    def _to_untrimmed_dict(self, in_cls = None) -> dict:
        untrimmed = {
            'className': self.class_name,
            'color': self.color,
            'cursor': self.cursor,
            'lineColor': self.line_color,
            'lineWidth': self.line_width,
            'zIndex': self.z_index
        }

        return untrimmed


class GuideBox(HighchartsMeta):
    """Style options for the guide box. The guide box has one state by default, the
    ``default`` state."""

    def __init__(self, **kwargs):
        self._default = None
        self.default = kwargs.get('default', None)

    @property
    def default(self) -> Optional[GuideBoxOptions]:
        """Style options for the guide box default state.

        :rtype: :class:`GuideBoxOptions` or :obj:`None <python:None>`
        """
        return self._default

    @default.setter
    @class_sensitive(GuideBoxOptions)
    def default(self, value):
        self._default = value

    @classmethod
    def _get_kwargs_from_dict(cls, as_dict):
        return {
            'default': as_dict.get('default', None)
        }

    def _to_untrimmed_dict(self, in_cls = None) -> dict:
        untrimmed = {
            'default': self.default
        }

        return untrimmed


class DragDropOptions(HighchartsMeta):
    """The draggable-points module allows points to be moved around or modified in the
    chart.

    In addition to the options mentioned under the dragDrop API structure, the module
    fires three (JavaScript) events:

      * ``point.dragStart``
      * ``point.drag``
      * ``point.drop``

    """

    def __init__(self, **kwargs):
        self._draggable_x = None
        self._draggable_y = None
        self._drag_handle = None
        self._drag_max_x = None
        self._drag_max_y = None
        self._drag_min_x = None
        self._drag_min_y = None
        self._drag_precision_x = None
        self._drag_precision_y = None
        self._drag_sensitivity = 2
        self._group_by = None
        self._guide_box = None
        self._live_redraw = True

        self.draggable_x = kwargs.get('draggable_x', None)
        self.draggable_y = kwargs.get('draggable_y', None)
        self.drag_handle = kwargs.get('drag_handle', None)
        self.drag_max_x = kwargs.get('drag_max_x', None)
        self.drag_max_y = kwargs.get('drag_max_y', None)
        self.drag_min_x = kwargs.get('drag_min_x', None)
        self.drag_min_y = kwargs.get('drag_min_y', None)
        self.drag_precision_x = kwargs.get('drag_precision_x', None)
        self.drag_precision_y = kwargs.get('drag_precision_y', None)
        self.drag_sensitivity = kwargs.get('drag_sensitivity', None)
        self.group_by = kwargs.get('group_by', None)
        self.guide_box = kwargs.get('guide_box', None)
        self.live_redraw = kwargs.get('live_redraw', None)

    @property
    def _dot_path(self) -> Optional[str]:
        """The dot-notation path to the options key for the current class.
        
        :rtype: :class:`str <python:str>` or :obj:`None <python:None>`
        """
        return 'plotOptions.series.dragDrop'

    @property
    def draggable_x(self) -> Optional[bool]:
        """If ``True``, enables dragging along the X dimension. Defaults to
        :obj:`None <python:None>`, which is equivalent to ``False``.

        :rtype: :class:`bool <python:bool>` or :obj:`None <python:None>`
        """
        return self._draggable_x

    @draggable_x.setter
    def draggable_x(self, value):
        if value is None:
            self._draggable_x = None
        else:
            self._draggable_x = bool(value)

    @property
    def draggable_y(self) -> Optional[bool]:
        """If ``True``, enables dragging along the Y dimension. Defaults to
        :obj:`None <python:None>`, which is equivalent to ``False``.

        .. warning::

          This is not supported for TreeGrid axes (the default axis type in Gantt charts).

        :rtype: :class:`bool <python:bool>` or :obj:`None <python:None>`
        """
        return self._draggable_y

    @draggable_y.setter
    def draggable_y(self, value):
        if value is None:
            self._draggable_y = None
        else:
            self._draggable_y = bool(value)

    @property
    def drag_handle(self) -> Optional[DragHandle]:
        """Options for the drag handles available in column series.

        :rtype: :class:`DragHandle` or :obj:`None <python:None>`
        """
        return self._drag_handle

    @drag_handle.setter
    @class_sensitive(DragHandle)
    def drag_handle(self, value):
        self._drag_handle = value

    @property
    def drag_max_x(self) -> Optional[int | float | Decimal]:
        """The maximum X value the points can be moved to. Defaults to
        :obj:`None <python:None>`.

        :rtype: numeric or :obj:`None <python:None>`
        """
        return self._drag_max_x

    @drag_max_x.setter
    def drag_max_x(self, value):
        self._drag_max_x = validators.numeric(value, allow_empty = True)

    @property
    def drag_max_y(self) -> Optional[int | float | Decimal]:
        """The maximum Y value the points can be moved to. Defaults to
        :obj:`None <python:None>`.

        :rtype: numeric or :obj:`None <python:None>`
        """
        return self._drag_max_y

    @drag_max_y.setter
    def drag_max_y(self, value):
        self._drag_max_y = validators.numeric(value, allow_empty = True)

    @property
    def drag_min_x(self) -> Optional[int | float | Decimal]:
        """The minimum X value the points can be moved to. Defaults to
        :obj:`None <python:None>`.

        :rtype: numeric or :obj:`None <python:None>`
        """
        return self._drag_min_x

    @drag_min_x.setter
    def drag_min_x(self, value):
        self._drag_min_x = validators.numeric(value, allow_empty = True)

    @property
    def drag_min_y(self) -> Optional[int | float | Decimal]:
        """The minimum Y value the points can be moved to. Defaults to
        :obj:`None <python:None>`.

        :rtype: numeric or :obj:`None <python:None>`
        """
        return self._drag_min_y

    @drag_min_y.setter
    def drag_min_y(self, value):
        self._drag_min_y = validators.numeric(value, allow_empty = True)

    @property
    def drag_precision_x(self) -> Optional[int | float | Decimal]:
        """The X precision value to drag to for this series. Defaults to
        :obj:`None <python:None>`, which is equivalent to disabling for most axes except
        for category axes where the default is ``1``.

        Set to ``0`` to disable for all axes.

        :rtype: numeric or :obj:`None <python:None>`
        """
        return self._drag_precision_x

    @drag_precision_x.setter
    def drag_precision_x(self, value):
        self._drag_precision_x = validators.numeric(value,
                                                    allow_empty = True,
                                                    minimum = 0)

    @property
    def drag_precision_y(self) -> Optional[int | float | Decimal]:
        """The Y precision value to drag to for this series. Defaults to
        :obj:`None <python:None>`, which is equivalent to disabling for most axes except
        for category axes where the default is ``1``.

        Set to ``0`` to disable for all axes.

        :rtype: numeric or :obj:`None <python:None>`
        """
        return self._drag_precision_y

    @drag_precision_y.setter
    def drag_precision_y(self, value):
        self._drag_precision_y = validators.numeric(value,
                                                    allow_empty = True,
                                                    minimum = 0)

    @property
    def drag_sensitivity(self) -> Optional[int | float | Decimal]:
        """The number of pixels to drag the pointer before it counts as a drag operation.
        Defaults to ``2``.

        .. note::

          This prevents drag/drop to fire when just clicking or selecting points.

        :rtype: numeric or :obj:`None <python:None>`
        """
        return self._drag_sensitivity

    @drag_sensitivity.setter
    def drag_sensitivity(self, value):
        self._drag_sensitivity = validators.numeric(value,
                                                    allow_empty = True,
                                                    minimum = 0)

    @property
    def group_by(self) -> Optional[str]:
        """Group the points by a property. Points with the same property value will be
        grouped together when moving. Defaults to :obj:`None <python:None>`.

        :rtype: :class:`str <python:str>` or :obj:`None <python:None>`
        """
        return self._group_by

    @group_by.setter
    def group_by(self, value):
        self._group_by = validators.string(value, allow_empty = True)

    @property
    def guide_box(self) -> Optional[GuideBox]:
        """Style options for the guide box. The guide box has one state by default, the
        ``default`` state.

        :rtype: :class:`GuideBox` or :obj:`None <python:None>`
        """
        return self._guide_box

    @guide_box.setter
    @class_sensitive(GuideBox)
    def guide_box(self, value):
        self._guide_box = value

    @property
    def live_redraw(self) -> Optional[bool]:
        """If ``True``, update points as they are dragged. If ``False``, a guide box is
        drawn to illustrate the new point size. Defaults to :obj:`None <python:None>`,
        which behaves as ``True``.

        :rtype: :class:`bool <python:bool>` or :obj:`None <python:None>`.
        """
        return self._live_redraw

    @live_redraw.setter
    def live_redraw(self, value):
        if value is None:
            self._live_redraw = None
        else:
            self._live_redraw = bool(value)

    @classmethod
    def _get_kwargs_from_dict(cls, as_dict):
        kwargs = {
            'draggable_x': as_dict.get('draggableX', None),
            'draggable_y': as_dict.get('draggableY', None),
            'drag_handle': as_dict.get('dragHandle', None),
            'drag_max_x': as_dict.get('dragMaxX', None),
            'drag_max_y': as_dict.get('dragMaxY', None),
            'drag_min_x': as_dict.get('dragMinX', None),
            'drag_min_y': as_dict.get('dragMinY', None),
            'drag_precision_x': as_dict.get('dragPrecisionX', None),
            'drag_precision_y': as_dict.get('dragPrecisionY', None),
            'drag_sensitivity': as_dict.get('dragSensitivity', None),
            'group_by': as_dict.get('groupBy', None),
            'guide_box': as_dict.get('guideBox', None),
            'live_redraw': as_dict.get('liveRedraw', None)
        }

        return kwargs

    def _to_untrimmed_dict(self, in_cls = None) -> dict:
        untrimmed = {
            'draggableX': self.draggable_x,
            'draggableY': self.draggable_y,
            'dragHandle': self.drag_handle,
            'dragMaxX': self.drag_max_x,
            'dragMaxY': self.drag_max_y,
            'dragMinX': self.drag_min_x,
            'dragMinY': self.drag_min_y,
            'dragPrecisionX': self.drag_precision_x,
            'dragPrecisionY': self.drag_precision_y,
            'dragSensitivity': self.drag_sensitivity,
            'groupBy': self.group_by,
            'guideBox': self.guide_box,
            'liveRedraw': self.live_redraw
        }

        return untrimmed


class HighLowDragDropOptions(DragDropOptions):
    """The draggable-points module allows points to be moved around or modified in the
    chart."""

    def __init__(self, **kwargs):
        self._draggable_high = None
        self._draggable_low = None

        self.draggable_high = kwargs.get('draggable_high', None)
        self.draggable_low = kwargs.get('draggable_low', None)

        super().__init__(**kwargs)

    @property
    def _dot_path(self) -> Optional[str]:
        """The dot-notation path to the options key for the current class.
        
        :rtype: :class:`str <python:str>` or :obj:`None <python:None>`
        """
        return 'plotOptions.boxplot.dragDrop'

    @property
    def draggable_high(self) -> Optional[bool]:
        """If ``True``, enables high value to be dragged individually. Defaults to
        :obj:`None <python:None>`, which is equivalent to ``True``.

        :rtype: :class:`bool <python:bool>` or :obj:`None <python:None>`
        """
        return self._draggable_high

    @draggable_high.setter
    def draggable_high(self, value):
        if value is None:
            self._draggable_high = None
        else:
            self._draggable_high = bool(value)

    @property
    def draggable_low(self) -> Optional[bool]:
        """If ``True``, enables the low value to be dragged individually. Defaults to
        :obj:`None <python:None>`, which is equivalent to ``True``.

        :rtype: :class:`bool <python:bool>` or :obj:`None <python:None>`
        """
        return self._draggable_low

    @draggable_low.setter
    def draggable_low(self, value):
        if value is None:
            self._draggable_low = None
        else:
            self._draggable_low = bool(value)

    @classmethod
    def _get_kwargs_from_dict(cls, as_dict):
        kwargs = {
            'draggable_x': as_dict.get('draggableX', None),
            'draggable_y': as_dict.get('draggableY', None),
            'drag_handle': as_dict.get('dragHandle', None),
            'drag_max_x': as_dict.get('dragMaxX', None),
            'drag_max_y': as_dict.get('dragMaxY', None),
            'drag_min_x': as_dict.get('dragMinX', None),
            'drag_min_y': as_dict.get('dragMinY', None),
            'drag_precision_x': as_dict.get('dragPrecisionX', None),
            'drag_precision_y': as_dict.get('dragPrecisionY', None),
            'drag_sensitivity': as_dict.get('dragSensitivity', None),
            'group_by': as_dict.get('groupBy', None),
            'guide_box': as_dict.get('guideBox', None),
            'live_redraw': as_dict.get('liveRedraw', None),

            'draggable_high': as_dict.get('draggableHigh', None),
            'draggable_low': as_dict.get('draggableLow', None),
        }

        return kwargs

    def _to_untrimmed_dict(self, in_cls = None) -> dict:
        untrimmed = {
            'draggableHigh': self.draggable_high,
            'draggableLow': self.draggable_low,
            'draggableX': self.draggable_x,
            'draggableY': self.draggable_y,
            'dragHandle': self.drag_handle,
            'dragMaxX': self.drag_max_x,
            'dragMaxY': self.drag_max_y,
            'dragMinX': self.drag_min_x,
            'dragMinY': self.drag_min_y,
            'dragPrecisionX': self.drag_precision_x,
            'dragPrecisionY': self.drag_precision_y,
            'dragSensitivity': self.drag_sensitivity,
            'groupBy': self.group_by,
            'guideBox': self.guide_box,
            'liveRedraw': self.live_redraw
        }

        return untrimmed


class BoxPlotDragDropOptions(HighLowDragDropOptions):
    """The draggable-points module allows points to be moved around or modified in the
    chart."""

    def __init__(self, **kwargs):
        self._draggable_high = None
        self._draggable_low = None
        self._draggable_q1 = None
        self._draggable_q3 = None

        self.draggable_high = kwargs.get('draggable_high', None)
        self.draggable_low = kwargs.get('draggable_low', None)
        self.draggable_q1 = kwargs.get('draggable_q1', None)
        self.draggable_q3 = kwargs.get('draggable_q3', None)

        super().__init__(**kwargs)

    @property
    def _dot_path(self) -> Optional[str]:
        """The dot-notation path to the options key for the current class.
        
        :rtype: :class:`str <python:str>` or :obj:`None <python:None>`
        """
        return 'plotOptions.boxplot.dragDrop'

    @property
    def draggable_q1(self) -> Optional[bool]:
        """If ``True``, enables the Q1 value to be dragged individually. Defaults to
        :obj:`None <python:None>`, which is equivalent to ``True``.

        :rtype: :class:`bool <python:bool>` or :obj:`None <python:None>`
        """
        return self._draggable_q1

    @draggable_q1.setter
    def draggable_q1(self, value):
        if value is None:
            self._draggable_q1 = None
        else:
            self._draggable_q1 = bool(value)

    @property
    def draggable_q3(self) -> Optional[bool]:
        """If ``True``, enables the Q3 value to be dragged individually. Defaults to
        :obj:`None <python:None>`, which is equivalent to ``True``.

        :rtype: :class:`bool <python:bool>` or :obj:`None <python:None>`
        """
        return self._draggable_q3

    @draggable_q3.setter
    def draggable_q3(self, value):
        if value is None:
            self._draggable_q3 = None
        else:
            self._draggable_q3 = bool(value)

    @classmethod
    def _get_kwargs_from_dict(cls, as_dict):
        kwargs = {
            'draggable_x': as_dict.get('draggableX', None),
            'draggable_y': as_dict.get('draggableY', None),
            'drag_handle': as_dict.get('dragHandle', None),
            'drag_max_x': as_dict.get('dragMaxX', None),
            'drag_max_y': as_dict.get('dragMaxY', None),
            'drag_min_x': as_dict.get('dragMinX', None),
            'drag_min_y': as_dict.get('dragMinY', None),
            'drag_precision_x': as_dict.get('dragPrecisionX', None),
            'drag_precision_y': as_dict.get('dragPrecisionY', None),
            'drag_sensitivity': as_dict.get('dragSensitivity', None),
            'group_by': as_dict.get('groupBy', None),
            'guide_box': as_dict.get('guideBox', None),
            'live_redraw': as_dict.get('liveRedraw', None),

            'draggable_high': as_dict.get('draggableHigh', None),
            'draggable_low': as_dict.get('draggableLow', None),
            'draggable_q1': as_dict.get('draggableQ1', None),
            'draggable_q3': as_dict.get('draggableQ3', None)
        }

        return kwargs

    def _to_untrimmed_dict(self, in_cls = None) -> dict:
        untrimmed = {
            'draggableX': self.draggable_x,
            'draggableY': self.draggable_y,
            'dragHandle': self.drag_handle,
            'dragMaxX': self.drag_max_x,
            'dragMaxY': self.drag_max_y,
            'dragMinX': self.drag_min_x,
            'dragMinY': self.drag_min_y,
            'dragPrecisionX': self.drag_precision_x,
            'dragPrecisionY': self.drag_precision_y,
            'dragSensitivity': self.drag_sensitivity,
            'groupBy': self.group_by,
            'guideBox': self.guide_box,
            'liveRedraw': self.live_redraw,

            'draggableHigh': self.draggable_high,
            'draggableLow': self.draggable_low,
            'draggableQ1': self.draggable_q1,
            'draggableQ3': self.draggable_q3
        }

        return untrimmed


class BulletDragDropOptions(DragDropOptions):
    """The draggable-points module allows points to be moved around or modified in the
    chart."""

    def __init__(self, **kwargs):
        self._draggable_target = None

        self.draggable_target = kwargs.get('draggable_target', None)

        super().__init__(**kwargs)

    @property
    def _dot_path(self) -> Optional[str]:
        """The dot-notation path to the options key for the current class.
        
        :rtype: :class:`str <python:str>` or :obj:`None <python:None>`
        """
        return 'plotOptions.bullet.dragDrop'

    @property
    def draggable_target(self) -> Optional[bool]:
        """If ``True``, enables the target to be dragged individually. Defaults to
        :obj:`None <python:None>`, which is equivalent to ``True``.

        :rtype: :class:`bool <python:bool>` or :obj:`None <python:None>`
        """
        return self._draggable_target

    @draggable_target.setter
    def draggable_target(self, value):
        if value is None:
            self._draggable_target = None
        else:
            self._draggable_target = bool(value)

    @classmethod
    def _get_kwargs_from_dict(cls, as_dict):
        kwargs = {
            'draggable_x': as_dict.get('draggableX', None),
            'draggable_y': as_dict.get('draggableY', None),
            'drag_handle': as_dict.get('dragHandle', None),
            'drag_max_x': as_dict.get('dragMaxX', None),
            'drag_max_y': as_dict.get('dragMaxY', None),
            'drag_min_x': as_dict.get('dragMinX', None),
            'drag_min_y': as_dict.get('dragMinY', None),
            'drag_precision_x': as_dict.get('dragPrecisionX', None),
            'drag_precision_y': as_dict.get('dragPrecisionY', None),
            'drag_sensitivity': as_dict.get('dragSensitivity', None),
            'group_by': as_dict.get('groupBy', None),
            'guide_box': as_dict.get('guideBox', None),
            'live_redraw': as_dict.get('liveRedraw', None),

            'draggable_target': as_dict.get('draggableTarget', None)
        }

        return kwargs

    def _to_untrimmed_dict(self, in_cls = None) -> dict:
        untrimmed = {
            'draggableX': self.draggable_x,
            'draggableY': self.draggable_y,
            'dragHandle': self.drag_handle,
            'dragMaxX': self.drag_max_x,
            'dragMaxY': self.drag_max_y,
            'dragMinX': self.drag_min_x,
            'dragMinY': self.drag_min_y,
            'dragPrecisionX': self.drag_precision_x,
            'dragPrecisionY': self.drag_precision_y,
            'dragSensitivity': self.drag_sensitivity,
            'groupBy': self.group_by,
            'guideBox': self.guide_box,
            'liveRedraw': self.live_redraw,

            'draggableTarget': self.draggable_target
        }

        return untrimmed



import datetime
import string
from typing import Optional
from decimal import Decimal
from collections import UserDict

from validator_collection import checkers, validators
import esprima
from esprima.error_handler import Error as ParseError

from highcharts_core import constants, errors


def serialize_to_js_literal(item, encoding = 'utf-8') -> Optional[str]:
    """Convert ``item`` to the contents of a JavaScript object literal code snippet.

    :param item: A value that is to be converted into a JS object literal notation value.

    :param encoding: The character encoding to apply to the resulting object. Defaults
      to ``'utf-8'``.
    :type encoding: :class:`str <python:str>`

    :returns: A JavaScript object literal code snippet, expressed as a string. Or
      :obj:`None <python:None>` if ``item`` is not serializable.
    :rtype: :class:`str <python:str>` or :obj:`None <python:None>`
    """
    if checkers.is_iterable(item, forbid_literals = (str, bytes, dict, UserDict)):
        return [serialize_to_js_literal(x, encoding = encoding) for x in item]
    elif hasattr(item, 'to_js_literal'):
        return item.to_js_literal(encoding = encoding)
    elif isinstance(item, constants.EnforcedNullType) or item == 'null':
        return constants.EnforcedNull
    elif isinstance(item, bool):
        return item
    elif checkers.is_string(item):
        return_value = item.replace("'", "\\'")
        return return_value
    elif checkers.is_numeric(item) and not isinstance(item, Decimal):
        return item
    elif isinstance(item, Decimal):
        return float(item)
    elif checkers.is_type(item, ('CallbackFunction')):
        return str(item)
    elif checkers.is_type(item, (dict, UserDict)):
        as_dict = {}
        for key in item:
            as_dict[key] = serialize_to_js_literal(item[key], encoding = encoding)
        return str(as_dict)
    elif checkers.is_datetime(item):
        if not item.tzinfo:
            item = item.replace(tzinfo = datetime.timezone.utc)
        return item.timestamp()*1000
    elif checkers.is_date(item):
        return f'Date.UTC({item.year}, {item.month - 1}, {item.day})'
    elif checkers.is_time(item):
        return item.isoformat()
    elif item is None:
        return None

    return None


def is_js_object(as_str):
    """Determine whether ``as_str`` is a JavaScript object.
    
    :param as_str: The string to evaluate.
    :type as_str: :class:`str <python:str>`
    
    :returns: ``True`` if ``as_str`` is a JavaScript function. ``False`` if not.
    :rtype: :class:`bool <python:bool>`
    """
    expression_item = f'const testName = {as_str}'
    try:
        parsed = esprima.parseScript(expression_item)
    except ParseError:
        try:
            parsed = esprima.parseModule(expression_item)
        except ParseError:
            return False

    body = parsed.body
    if not body:
        return False

    first_item = body[0]
    if first_item.type != 'VariableDeclaration':
        return False

    init = first_item.declarations[0].init
    if not init:
        return False
    if init.type in ('ObjectExpression'):
        return True

    return False


def attempt_variable_declaration(as_str):
    """Attempt to coerce ``as_str`` to a JavaScript variable declaration form.

    :param as_str: The string to evaluate.
    :type as_str: :class:`str <python:str>`

    :returns: ``True`` if ``as_str`` is a JavaScript function. ``False`` if not.
    :rtype: :class:`bool <python:bool>`
    """
    expression_item = f'const testName = {as_str}'
    try:
        parsed = esprima.parseScript(expression_item)
    except ParseError:
        try:
            parsed = esprima.parseModule(expression_item)
        except ParseError:
            return False

    body = parsed.body
    if not body:
        return False

    first_item = body[0]
    if first_item.type != 'VariableDeclaration':
        return False

    init = first_item.declarations[0].init
    if not init:
        return False
    if init.type in ('FunctionExpression', 'ArrowFunctionExpression', 'ClassExpression'):
        return True
    elif init.type == 'NewExpression':
        callee = init.callee
        if not callee:
            return False
        if callee.name == 'Function':
            return True

    return False


def is_js_function_or_class(as_str) -> bool:
    """Determine whether ``as_str`` is a JavaScript function or not.

    :param as_str: The string to evaluate.
    :type as_str: :class:`str <python:str>`

    :returns: ``True`` if ``as_str`` is a JavaScript function. ``False`` if not.
    :rtype: :class:`bool <python:bool>`
    """
    if not checkers.is_string(as_str):
        return False

    try:
        parsed = esprima.parseScript(as_str)
    except ParseError:
        try:
            parsed = esprima.parseModule(as_str)
        except ParseError:
            if as_str.startswith('function') is False:
                return False
            else:
                return attempt_variable_declaration(as_str)

    body = parsed.body
    if not body:
        return False

    first_item = body[0]
    if first_item.type in ('FunctionDeclaration', 'ClassDeclaration'):
        return True
    elif as_str.startswith('function') or as_str.startswith('class'):
        return attempt_variable_declaration(as_str)
    elif first_item.type == 'VariableDeclaration':
        init = first_item.declarations[0].init
        if not init:
            return False
        if init.type in ('FunctionExpression', 'ArrowFunctionExpression',
                         'ClassExpression'):
            return True
        elif init.type == 'NewExpression':
            callee = init.callee
            if not callee:
                return False
            if callee.name == 'Function':
                return True

    return False


def get_js_literal(item) -> str:
    """Convert the value of ``item`` into a JavaScript literal string.

    :returns: The JavaScript literal string.
    :rtype: :class:`str <python:str>`
    """
    as_str = ''
    if checkers.is_iterable(item, forbid_literals = (str, bytes, dict, UserDict)):
        subitems = [get_js_literal(x) for x in item]
        as_str += '['
        subitem_counter = 0
        for subitem in subitems:
            subitem_counter += 1
            if subitem == 'None':
                subitem = 'null'
            as_str += f"""{subitem}"""
            if subitem_counter < len(subitems):
                as_str += ',\n'
        as_str += ']'
    elif checkers.is_string(item):
        if (item.startswith('[') or item.startswith('Date')) and item != 'Date':
            as_str += f"""{item}"""
        elif item.startswith('{') and item.endswith('}'):
            if is_js_object(item):
                as_str += f"""{item}"""
            elif "'" in item:
                item = item.replace("'", "\\'")
                as_str += f'"{item}"'
            else:
                as_str += f"'{item}'"
        elif item in string.whitespace:
            as_str += f"""`{item}`"""
        elif item.startswith == 'HCP: REPLACE-WITH-':
            item_str = item.replace('HCP: REPLACE-WITH-', '')
            as_str += f"""{item_str}"""
        elif not is_js_function_or_class(item):
            as_str += f"""'{item}'"""
        else:
            as_str += f"""{item}"""
    elif item == constants.EnforcedNull:
        as_str += """null"""
    elif item is True:
        as_str += """true"""
    elif item is False:
        as_str += """false"""
    else:
        as_str += f"""{item}"""

    return as_str


def assemble_js_literal(as_dict, keys_as_strings = False) -> Optional[str]:
    """Convert ``as_dict`` into a JavaScript object literal string.

    :param as_dict: A :class:`dict <python:dict>` representation of the JavaScript object.
    :type as_dict: :class:`dict <pythoN:dict>`

    :param keys_as_strings: if ``True``, will return the keys as string values (wrapped
      in quotation marks). If ``False``, will return the keys as object literals. Defaults
      to ``False``.
    :type keys_as_strings: :class:`bool <python:bool>`

    :returns: The JavaScript object literal representation of ``as_dict``.
    :rtype: :class:`str <python:str>` or :obj:`None <python:None>`
    """
    as_dict = validators.dict(as_dict, allow_empty = True)
    if not as_dict:
        return None

    as_str = ''
    keys = len(as_dict)
    current_key = 0
    ended_on_None = False

    for key in as_dict:
        current_key += 1
        item = as_dict[key]
        if item is None:
            keys -= 1
            ended_on_None = True
            continue

        if keys_as_strings:
            as_str += f"""  '{key}': """
        else:
            as_str += f"""  {key}: """

        as_str += get_js_literal(item)

        if current_key < keys:
            as_str += ',\n'
        else:
            ended_on_None = False
            as_str += '\n'

    if not as_str:
        return None
    elif ended_on_None:
        as_str = as_str[:-2]
        as_str += '\n'

    as_str = '{\n' + as_str + '}'

    return as_str


def convert_js_literal_to_python(literal_definition, original_str: None):
    """Convert a :class:`esprima.nodes.Literal` object into a Python literal.

    .. note::

      The :class:`esprima.nodes.Property` objects are available in the ``value`` sub-item.

    :rtype: Python objcet
    """
    if not checkers.is_type(literal_definition, ('Literal', 'Identifier')):
        raise errors.HighchartsParseError(f'literal_definition should contain a '
                                          f'Literal or Identifier instance. Received: '
                                          f'{literal_definition.__class__.__name__}')

    if literal_definition.raw == 'null':
        return constants.EnforcedNull
    elif literal_definition.name == 'undefined':
        return None
    elif literal_definition.value is not None:
        return literal_definition.value
    else:
        raise errors.HighchartsParseError('unable to find a literal, array, or object '
                                          'definition')


def convert_js_property_to_python(property_definition, original_str = None):
    """Convert a :class:`esprima.nodes.Property` object into a Python literal.

    .. note::

      The :class:`esprima.nodes.Property` objects are available in the ``value`` sub-item.

    """
    from highcharts_core.utility_classes.javascript_functions import CallbackFunction, \
        JavaScriptClass

    if not checkers.is_type(property_definition, 'Property'):
        raise errors.HighchartsParseError(f'property_definition should contain a '
                                          f'Property instance. Received: '
                                          f'{property_definition.__class__.__name__}')

    if property_definition.value.type == 'Literal':
        if property_definition.value.raw == 'null':
            return constants.EnforcedNull
        elif property_definition.value.name == 'undefined':
            return None
        elif property_definition.value is not None:
            return property_definition.value.value
        else:
            raise errors.HighchartsParseError(f'unexpected parse error when '
                                              f'interpreting:\n{property_definition}')
    elif property_definition.value.type == 'TemplateLiteral':
        template_elements = property_definition.value.quasis
        if template_elements:
            element = template_elements[0]
            if element.type == 'TemplateElement':
                return element.value.cooked
            else:
                raise errors.HighchartsParseError('unable to properly parse a '
                                                  'TemplateLiteral. Specifically could '
                                                  'not find a TemplateElement where one '
                                                  'was expected.')
    elif property_definition.value.type == 'UnaryExpression':
        property_value = property_definition.value
        operator = property_value.operator
        if operator == '-':
            multiple = -1
        else:
            multiple = 1
        argument_type = property_value.argument.type
        if argument_type not in ['Literal']:
            raise errors.HighchartsParseError(f'unable to find a Literal value within'
                                              f'a Unary expression. Found: '
                                              f'{argument_type}')
        value = property_value.argument.value
        if checkers.is_numeric(value):
            return value * multiple
        else:
            return value

    elif property_definition.value.type == 'Identifier':
        if property_definition.value.name == 'undefined':
            return None
        else:
            raise errors.HighchartsParseError(f'unexpected parse error when '
                                              f'interpreting:\n{property_definition}')
    elif property_definition.value.type == 'ArrayExpression':
        return [convert_js_to_python(x, original_str)
                for x in property_definition.value.elements]
    elif property_definition.value.type == 'ObjectExpression':
        as_dict = {}
        key_value_pairs = get_key_value_pairs(property_definition.value.properties,
                                              original_str)
        for pair in key_value_pairs:
            as_dict[pair[0]] = pair[1]

        return as_dict
    elif property_definition.value.type == 'FunctionExpression':
        return CallbackFunction._convert_from_js_ast(property_definition, original_str)
    elif property_definition.value.type == 'ClassExpression':
        return JavaScriptClass._convert_from_js_ast(property_definition.value,
                                                    original_str)
    elif property_definition.value.type == 'CallExpression':
        expression = property_definition.value
        try:
            callee_obj = expression.callee.object.name
        except AttributeError:
            try:
                callee_obj = expression.callee.name
            except AttributeError:
                raise errors.HighchartsParseError('unable to parse the JS Call '
                                                  'Expression')
        call_arguments = [x.value for x in expression.arguments]
        if callee_obj == 'Date':
            if len(call_arguments) == 1:
                return validators.datetime(call_arguments[0])
            elif len(call_arguments) == 2:
                return datetime.date(year = call_arguments[0],
                                     month = call_arguments[1] + 1)
            elif len(call_arguments) == 3:
                return datetime.date(year = call_arguments[0],
                                     month = call_arguments[1] + 1,
                                     day = call_arguments[2])
            elif len(call_arguments) == 4:
                return datetime.datetime(year = call_arguments[0],
                                         month = call_arguments[1] + 1,
                                         day = call_arguments[2],
                                         hour = call_arguments[3])
            elif len(call_arguments) == 5:
                return datetime.datetime(year = call_arguments[0],
                                         month = call_arguments[1] + 1,
                                         day = call_arguments[2],
                                         hour = call_arguments[3],
                                         minute = call_arguments[4])
            elif len(call_arguments) == 6:
                return datetime.datetime(year = call_arguments[0],
                                         month = call_arguments[1] + 1,
                                         day = call_arguments[2],
                                         hour = call_arguments[3],
                                         minute = call_arguments[4],
                                         second = call_arguments[5])
            elif len(call_arguments) == 7:
                return datetime.datetime(year = call_arguments[0],
                                         month = call_arguments[1] + 1,
                                         day = call_arguments[2],
                                         hour = call_arguments[3],
                                         minute = call_arguments[4],
                                         second = call_arguments[5],
                                         microsecond = call_arguments[6])
            else:
                raise errors.HighchartsParseError('failed to parse the Date() '
                                                  'constructor from the JS literal')

    else:
        raise errors.HighchartsParseError('unable to find a literal, array, or object '
                                          'definition')


def convert_js_to_python(javascript, original_str = None):
    """Convert a :class:`esprima.nodes.Property` object into a Python literal.

    .. note::

      The :class:`esprima.nodes.Property` objects are available in the ``value`` sub-item.

    """
    if javascript.type not in ('Property',
                               'Literal',
                               'ObjectExpression',
                               'ArrayExpression',
                               'UnaryExpression',
                               'FunctionExpression'):
        raise errors.HighchartsParseError(f'javascript should contain a '
                                          f'Property, Literal, ObjectExpression, '
                                          f'ArrayExpression, UnaryExpression, or '
                                          f'FunctionExpression instance. Received: '
                                          f'{javascript.type}')

    if checkers.is_type(javascript, 'Property'):
        return convert_js_property_to_python(javascript, original_str)
    elif checkers.is_type(javascript, 'ObjectExpression'):
        as_dict = {}
        key_value_pairs = get_key_value_pairs(javascript.properties, original_str)
        for pair in key_value_pairs:
            as_dict[pair[0]] = pair[1]

        return as_dict
    elif checkers.is_type(javascript, 'ArrayExpression'):
        return [convert_js_to_python(x, original_str)
                for x in javascript.elements]
    elif checkers.is_type(javascript, 'UnaryExpression'):
        is_negative = javascript.operator == '-'
        if is_negative:
            multiple = -1
        converted_value = convert_js_literal_to_python(javascript.argument, original_str)
        return converted_value * multiple
    elif checkers.is_type(javascript, 'FunctionExpression'):
        from highcharts_core.utility_classes.javascript_functions import CallbackFunction

        start_char = javascript.range[0]
        end_char = javascript.range[1]
        function_as_str = original_str[start_char:end_char]
        return CallbackFunction.from_js_literal(function_as_str)
    else:
        return convert_js_literal_to_python(javascript, original_str)


def get_key_value_pairs(properties, original_str):
    """Return the key and value pairs for properties defined in ``properties``.

    :param properties: The definition of a JavaScript object Property.
    :type properties: :class:`list <python:list>` of :class:`esprima.nodes.Property`
      instances

    :rtype: :class:`list <python:list>` of :class:`tuples <python:tuple>` of
      :class:`str <python:str>` and any value
    """
    key_value_pairs = []
    for item in properties:
        if not checkers.is_type(item, 'Property'):
            raise errors.HighchartsParseError(f'properties should contain a list of '
                                              f'Property instances. Received: '
                                              f'{item.__class__.__name__}')
        key = item.key.name or item.key.value
        if not key:
            raise errors.HighchartsMissingKeyError('property was not formed correctly, '
                                                   'specifically missing a key')
        value = convert_js_to_python(item, original_str)
        pair = (key, value)
        key_value_pairs.append(pair)

    return key_value_pairs



"""Tests for ``highcharts.utility_classes.ast``."""

import pytest

from json.decoder import JSONDecodeError
from copy import deepcopy

from highcharts_core.utility_classes.ast import AttributeObject as cls
from highcharts_core.utility_classes.ast import ASTNode as cls2
from highcharts_core.utility_classes.ast import ASTMap as cls3
from highcharts_core.utility_classes.ast import TextPath as cls4

from highcharts_core import errors
from tests.fixtures import input_files, check_input_file, to_camelCase, to_js_dict, \
    Class__init__, Class__to_untrimmed_dict, Class_from_dict, Class_to_dict, \
    Class_from_js_literal

STANDARD_PARAMS_cls = [
    ({}, None),
    ({
        'stroke': '#ff0000',
        'stroke-width': 2,
        'rotation': 45,
        'd': ['M', 10, 10, 'L', 30, 30, 'z']
    }, None),
]


@pytest.mark.parametrize('kwargs, error', STANDARD_PARAMS_cls)
def test_AttributeObject__init__(kwargs, error):
    kwargs_copy = deepcopy(kwargs)
    if not error:
        result = cls(**kwargs)
        print(kwargs_copy)
        assert result is not None
        assert isinstance(result, cls) is True
        for key in kwargs_copy:
            if isinstance(kwargs_copy[key], str) and kwargs[key].startswith('function'):
                continue
            if isinstance(kwargs_copy[key], str) and kwargs[key].startswith('class'):
                continue

            assert kwargs_copy[key] == result.get(key)
    else:
        with pytest.raises(error):
            result = cls(**kwargs)


@pytest.mark.parametrize('kwargs, error', STANDARD_PARAMS_cls)
def test_AttributeObject__to_untrimmed_dict(kwargs, error):
    Class__to_untrimmed_dict(cls, kwargs, error)


@pytest.mark.parametrize('kwargs, error',  STANDARD_PARAMS_cls)
def test_AttributeObject_from_dict(kwargs, error):
    as_dict = to_js_dict(deepcopy(kwargs))

    if not error:
        instance = cls.from_dict(as_dict)
        assert instance is not None
        assert isinstance(instance, cls) is True
        for key in kwargs:
            assert kwargs.get(key) == instance.get(key)
    else:
        with pytest.raises(error):
            instance = cls.from_dict(as_dict)


@pytest.mark.parametrize('kwargs, error',  STANDARD_PARAMS_cls)
def test_AttributeObject_to_dict(kwargs, error):
    Class_to_dict(cls, kwargs, error)


STANDARD_PARAMS_cls2 = [
    ({}, None),

]


@pytest.mark.parametrize('kwargs, error', STANDARD_PARAMS_cls2)
def test_ASTNode__init__(kwargs, error):
    Class__init__(cls, kwargs, error)


@pytest.mark.parametrize('kwargs, error', STANDARD_PARAMS_cls2)
def test_ASTNode__to_untrimmed_dict(kwargs, error):
    Class__to_untrimmed_dict(cls, kwargs, error)


@pytest.mark.parametrize('kwargs, error',  STANDARD_PARAMS_cls2)
def test_ASTNode_from_dict(kwargs, error):
    Class_from_dict(cls, kwargs, error)


@pytest.mark.parametrize('kwargs, error',  STANDARD_PARAMS_cls2)
def test_ASTNode_to_dict(kwargs, error):
    Class_to_dict(cls, kwargs, error)


STANDARD_PARAMS_cls3 = [
    ({}, None),
]


@pytest.mark.parametrize('kwargs, error', STANDARD_PARAMS_cls3)
def test_ASTMap__init__(kwargs, error):
    kwargs_copy = deepcopy(kwargs)
    if not error:
        result = cls(**kwargs)
        print(kwargs_copy)
        assert result is not None
        assert isinstance(result, cls) is True
        for key in kwargs_copy:
            if isinstance(kwargs_copy[key], str) and kwargs[key].startswith('function'):
                continue
            if isinstance(kwargs_copy[key], str) and kwargs[key].startswith('class'):
                continue

            assert kwargs_copy[key] == result.get(key)
    else:
        with pytest.raises(error):
            result = cls(**kwargs)


@pytest.mark.parametrize('kwargs, error', STANDARD_PARAMS_cls3)
def test_ASTMap__to_untrimmed_dict(kwargs, error):
    Class__to_untrimmed_dict(cls, kwargs, error)


@pytest.mark.parametrize('kwargs, error',  STANDARD_PARAMS_cls3)
def test_ASTMap_from_dict(kwargs, error):
    as_dict = to_js_dict(deepcopy(kwargs))

    if not error:
        instance = cls.from_dict(as_dict)
        assert instance is not None
        assert isinstance(instance, cls) is True
        for key in kwargs:
            assert kwargs.get(key) == instance.get(key)
    else:
        with pytest.raises(error):
            instance = cls.from_dict(as_dict)


@pytest.mark.parametrize('kwargs, error',  STANDARD_PARAMS_cls3)
def test_ASTMap_to_dict(kwargs, error):
    Class_to_dict(cls, kwargs, error)


STANDARD_PARAMS_cls4 = [
    ({}, None),

]


@pytest.mark.parametrize('kwargs, error', STANDARD_PARAMS_cls4)
def test_TextPath__init__(kwargs, error):
    Class__init__(cls, kwargs, error)


@pytest.mark.parametrize('kwargs, error', STANDARD_PARAMS_cls4)
def test_TextPath__to_untrimmed_dict(kwargs, error):
    Class__to_untrimmed_dict(cls, kwargs, error)


@pytest.mark.parametrize('kwargs, error',  STANDARD_PARAMS_cls4)
def test_TextPath_from_dict(kwargs, error):
    Class_from_dict(cls, kwargs, error)


@pytest.mark.parametrize('kwargs, error',  STANDARD_PARAMS_cls4)
def test_TextPath_to_dict(kwargs, error):
    Class_to_dict(cls, kwargs, error)



from typing import Optional

from validator_collection import validators

from highcharts_core import constants
from highcharts_core.decorators import class_sensitive
from highcharts_core.metaclasses import HighchartsMeta


class BoostDebug(HighchartsMeta):
    """Debugging options for boost. Useful for benchmarking, and general timing."""

    def __init__(self, **kwargs):
        self._show_skip_summary = None
        self._time_buffer_copy = None
        self._time_kd_tree = None
        self._time_rendering = None
        self._time_series_processing = None
        self._time_setup = None

        self.show_skip_summary = kwargs.get('show_skip_summary', None)
        self.time_buffer_copy = kwargs.get('time_buffer_copy', None)
        self.time_kd_tree = kwargs.get('time_kd_tree', None)
        self.time_rendering = kwargs.get('time_rendering', None)
        self.time_series_processing = kwargs.get('time_series_processing', None)
        self.time_setup = kwargs.get('time_setup', None)

    @property
    def show_skip_summary(self) -> Optional[bool]:
        """When ``True``, the number of points skipped in series processing is outputted.
        Defaults to ``False``.

        Points are skipped if they are closer than 1 pixel from each other.

        :returns: Flag indicating whether the skip summary will be shown.
        :rtype: :class:`bool <python:bool>` or :obj:`None <python:None>`
        """
        return self._show_skip_summary

    @show_skip_summary.setter
    def show_skip_summary(self, value):
        if value is None:
            self._show_skip_summary = None
        else:
            self._show_skip_summary = bool(value)

    @property
    def time_buffer_copy(self) -> Optional[bool]:
        """When ``True``, the time it takes for the SVG buffer copy to complete is
        outputted. Defaults to ``False``.

        :returns: Flag indicating whether the buffer copy timing will be shown.
        :rtype: :class:`bool <python:bool>` or :obj:`None <python:None>`
        """
        return self._time_buffer_copy

    @time_buffer_copy.setter
    def time_buffer_copy(self, value):
        if value is None:
            self._time_buffer_copy = None
        else:
            self._time_buffer_copy = bool(value)

    @property
    def time_kd_tree(self) -> Optional[bool]:
        """When ``True``, the time spent building the k-d tree used for markers, etc. will
        be rendered. Defaults to ``False``.

        .. note::

          Note that the k-d tree is built asynchronously, and runs post-rendering.
          Thus, it does not affect the performance of the rendering itself.

        :returns: Flag indicating whether the KD tree timing will be shown.
        :rtype: :class:`bool <python:bool>` or :obj:`None <python:None>`
        """
        return self._time_kd_tree

    @time_kd_tree.setter
    def time_kd_tree(self, value):
        if value is None:
            self._time_kd_tree = None
        else:
            self._time_kd_tree = bool(value)

    @property
    def time_rendering(self) -> Optional[bool]:
        """When ``True``, the time spent on actual rendering is outputted to the console.
        Defaults to ``False``.

        :returns: Flag indicating whether the rendering time will be shown.
        :rtype: :class:`bool <python:bool>` or :obj:`None <python:None>`
        """
        return self._time_rendering

    @time_rendering.setter
    def time_rendering(self, value):
        if value is None:
            self._time_rendering = None
        else:
            self._time_rendering = bool(value)

    @property
    def time_series_processing(self) -> Optional[bool]:
        """When ``True``, the time spent on transforming the series data to vertex buffers
        is outputted. Defaults to ``False``.

        :returns: Flag indicating whether the series processing time will be shown.
        :rtype: :class:`bool <python:bool>` or :obj:`None <python:None>`
        """
        return self._time_series_processing

    @time_series_processing.setter
    def time_series_processing(self, value):
        if value is None:
            self._time_series_processing = None
        else:
            self._time_series_processing = bool(value)

    @property
    def time_setup(self) -> Optional[bool]:
        """When ``True``, the te time spent on setting up the WebGL context, creating
        shaders, and textures is outputted. Defaults to ``False``.

        :returns: Flag indicating whether the setup time will be shown.
        :rtype: :class:`bool <python:bool>` or :obj:`None <python:None>`
        """
        return self._time_setup

    @time_setup.setter
    def time_setup(self, value):
        if value is None:
            self._time_setup = None
        else:
            self._time_setup = bool(value)

    @classmethod
    def _get_kwargs_from_dict(cls, as_dict):
        kwargs = {
            'show_skip_summary': as_dict.get('showSkipSummary', None),
            'time_buffer_copy': as_dict.get('timeBufferCopy', None),
            'time_kd_tree': as_dict.get('timeKDTree', None),
            'time_rendering': as_dict.get('timeRendering', None),
            'time_series_processing': as_dict.get('timeSeriesProcessing', None),
            'time_setup': as_dict.get('timeSetup', None),
        }

        return kwargs

    def _to_untrimmed_dict(self, in_cls = None) -> dict:
        return {
            'showSkipSummary': self.show_skip_summary,
            'timeBufferCopy': self.time_buffer_copy,
            'timeKDTree': self.time_kd_tree,
            'timeRendering': self.time_rendering,
            'timeSeriesProcessing': self.time_series_processing,
            'timeSetup': self.time_setup
        }


class Boost(HighchartsMeta):
    """Options for the Boost module.

    The Boost module allows certain series types to be rendered by WebGL instead of the
    default SVG. This allows hundreds of thousands of data points to be rendered in
    milliseconds. In addition to the WebGL rendering it saves time by skipping processing
    and inspection of the data wherever possible. This introduces some limitations to what
    features are available in boost mode. See
    `the docs <https://www.highcharts.com/docs/advanced-chart-features/boost-module>`_
    for details.

    .. note::

      In addition to the global :meth:`HighchartOptions.boost` property, each series has
      a :meth:`Series.boost_threshold` that defines when the boost should kick in.

    """

    def __init__(self, **kwargs):
        self._allow_force = None
        self._debug = None
        self._enabled = None
        self._pixel_ratio = None
        self._series_threshold = None
        self._use_gpu_translations = None
        self._use_preallocated = None

        self.allow_force = kwargs.get('allow_force', None)
        self.debug = kwargs.get('debug', None)
        self.enabled = kwargs.get('enabled', None)
        self.pixel_ratio = kwargs.get('pixel_ratio', None)
        self.series_threshold = kwargs.get('series_threshold', None)
        self.use_gpu_translations = kwargs.get('use_gpu_translations', None)
        self.use_preallocated = kwargs.get('use_preallocated', None)

    @property
    def _dot_path(self) -> Optional[str]:
        """The dot-notation path to the options key for the current class.
        
        :rtype: :class:`str <python:str>` or :obj:`None <python:None>`
        """
        return 'boost'

    @property
    def allow_force(self) -> Optional[bool]:
        """If ``True``, the whole chart will be boosted if one of the series crosses its
        threshold and all the series can be boosted. Defaults to ``True``.

        :returns: Flag indicating whether the entire chart can be boosted in response to
          one series being boosted.
        :rtype: class:`bool <python:bool>` or :obj:`None <python:None>`
        """
        return self._allow_force

    @allow_force.setter
    def allow_force(self, value):
        if value is None:
            self._allow_force = None
        else:
            self._allow_force = bool(value)

    @property
    def debug(self) -> Optional[BoostDebug]:
        """Debugging options for boost. Useful for benchmarking, and general timing.

        :rtype: :class:`BoostDebug` or :obj:`None <python:None>`
        """
        return self._debug

    @debug.setter
    @class_sensitive(BoostDebug)
    def debug(self, value):
        self._debug = value

    @property
    def enabled(self) -> Optional[bool]:
        """If ``True``, boost is enabled on the chart. If ``False``, boost is disabled.
        Defaults to ``True``.

        :returns: Flag indicating whether boost is enabled for the chart.
        :rtype: :class:`bool <python:bool>` or :obj:`None <python:None>`
        """
        return self._enabled

    @enabled.setter
    def enabled(self, value):
        if value is None:
            self._enabled = None
        else:
            self._enabled = bool(value)

    @property
    def pixel_ratio(self) -> Optional[int]:
        """The pixel ratio for the WebGL content. Defaults to
        ``1``.

        If ``0``, the ``window.devicePixelRatio`` is used. This ensures sharp graphics on
        high DPI displays like Apple's Retina, as well as when a page is zoomed.

        .. note::

          The default is left at ``1`` for now, as this
          is a new feature that has the potential to break existing setups. Over time,
          when it has been battle tested, the intention is to set it to ``0`` by default.

        .. hint::

          Another use case for this option is to set it to ``2`` in order to make
          exported and upscaled charts render sharp.

        .. warning::

          One limitation when using the ``pixel_ratio`` is that the line width of graphs
          is scaled down. Since the Boost module currently can only render ``1px`` line
          widths, it is scaled down to a thin ``0.5`` pixels on a Retina display.

        :rtype: :class:`int <python:int>` or :obj:`None <python:None>`
        """
        return self._pixel_ratio

    @pixel_ratio.setter
    def pixel_ratio(self, value):
        self._pixel_ratio = validators.integer(value,
                                               allow_empty = True,
                                               minimum = 0)

    @property
    def series_threshold(self) -> Optional[int]:
        """Set the series threshold for when the boost should kick in globally. Defaults
        to ``50``.

        Setting to e.g. ``20`` will cause the whole chart to enter boost mode if there are
        20 or more series active. When the chart is in boost mode, every series in it will
        be rendered to a common canvas. This offers a significant speed improvment in
        charts with a very high amount of series.

        :rtype: :class:`int <python:int>` or :obj:`None <python:None>`
        """
        return self._series_threshold

    @series_threshold.setter
    def series_threshold(self, value):
        self._series_threshold = validators.integer(value, allow_empty = True)

    @property
    def use_gpu_translations(self) -> Optional[bool]:
        """If ``True``, enables GPU translations. GPU translations are faster than doing
        the translation in JavaScript. Defaults to ``False``.

        .. warning::

          This option may cause rendering issues with certain datasets. Namely, if your
          dataset has large numbers with small increments (such as timestamps), it won't
          work correctly. This is due to floating point precission.

        :returns: Flag indicating whether to use GPU translations.
        :rtype: :class:`bool <python:bool>` or :obj:`None <python:None.`
        """
        return self._use_gpu_translations

    @use_gpu_translations.setter
    def use_gpu_translations(self, value):
        if value is None:
            self._use_gpu_translations = None
        else:
            self._use_gpu_translations = bool(value)

    @property
    def use_preallocated(self) -> Optional[bool]:
        """If ``True``, enables the pre-allocation of vertex buffers. Defaults to
        ``False``.

        Enabling this will make it so that the binary data arrays required for storing the
        series data will be allocated prior to transforming the data to a WebGL-compatible
        format.

        .. warning::

          Enabling this feature saves a copy operation on the order of O(n) and so is
          significantly more performant. However, this is currently an experimental
          option, and may cause visual artifacts with some datasets.

          As such, care should be taken when using this setting to make sure that it
          doesn't cause any rendering glitches with the given use-case.

        :returns: Flag indicating whether pre-allocation of vertex buffers is enabled.
        :rtype: :class:`bool <python:bool>` or :obj:`None <python:None>`
        """
        return self._use_preallocated

    @use_preallocated.setter
    def use_preallocated(self, value):
        if value is None:
            self._use_preallocated = None
        else:
            self._use_preallocated = bool(value)

    @classmethod
    def _get_kwargs_from_dict(cls, as_dict):
        kwargs = {
            'allow_force': as_dict.get('allowForce', None),
            'debug': as_dict.get('debug', None),
            'enabled': as_dict.get('enabled', None),
            'pixel_ratio': as_dict.get('pixelRatio', None),
            'series_threshold': as_dict.get('seriesThreshold', None),
            'use_gpu_translations': as_dict.get('useGPUTranslations', None),
            'use_preallocated': as_dict.get('usePreallocated', None),
        }

        return kwargs

    def _to_untrimmed_dict(self, in_cls = None) -> dict:
        untrimmed = {
            'allowForce': self.allow_force,
            'debug': self.debug,
            'enabled': self.enabled,
            'pixelRatio': self.pixel_ratio,
            'seriesThreshold': self.series_threshold,
            'useGPUTranslations': self.use_gpu_translations,
            'usePreallocated': self.use_preallocated
        }

        return untrimmed



from typing import Optional

from validator_collection import validators

from highcharts_core import constants
from highcharts_core.metaclasses import HighchartsMeta


class ExportingLanguageOptions(HighchartsMeta):
    """Exporting menu format strings for use in the accessibility module."""

    def __init__(self, **kwargs):
        self._chart_menu_label = None
        self._menu_button_label = None

        self.chart_menu_label = kwargs.get('chart_menu_label', None)
        self.menu_button_label = kwargs.get('menu_button_label', None)

    @property
    def chart_menu_label(self) -> Optional[str]:
        """Defaults to ``'Chart menu'``.

        :rtype: :class:`str <python:str>` or :obj:`None <python:None>`
        """
        return self._chart_menu_label

    @chart_menu_label.setter
    def chart_menu_label(self, value):
        self._chart_menu_label = validators.string(value, allow_empty = True)

    @property
    def menu_button_label(self) -> Optional[str]:
        """Defaults to ``'View chart menu, {chartTitle}'``.

        :rtype: :class:`str <python:str>` or :obj:`None <python:None>`
        """
        return self._menu_button_label

    @menu_button_label.setter
    def menu_button_label(self, value):
        self._menu_button_label = validators.string(value, allow_empty = True)

    @classmethod
    def _get_kwargs_from_dict(cls, as_dict):
        kwargs = {
            'chart_menu_label': as_dict.get('chartMenuLabel', None),
            'menu_button_label': as_dict.get('menuButtonLabel', None),
        }

        return kwargs

    def _to_untrimmed_dict(self, in_cls = None) -> dict:
        untrimmed = {
            'chartMenuLabel': self.chart_menu_label,
            'menuButtonLabel': self.menu_button_label
        }

        return untrimmed



from typing import Optional
from decimal import Decimal

from validator_collection import validators

from highcharts_core.decorators import class_sensitive
from highcharts_core.metaclasses import HighchartsMeta


class FocusBorderStyle(HighchartsMeta):
    """Style options for the focus border drawn around elements while navigating
    through them.

    .. warning::

      Some browsers, in addition, draw their own borders for focused elements.
      These automatic borders can not be styled by Highcharts.

    In styled mode, the border is given the ``.highcharts-focus-border`` class.
    """

    def __init__(self, **kwargs):
        self._border_radius = None
        self._color = None
        self._line_width = None

        self.border_radius = kwargs.get('border_radius', None)
        self.color = kwargs.get('color', None)
        self.line_width = kwargs.get('line_width', None)

    @property
    def border_radius(self) -> Optional[int | float | Decimal]:
        """The border radius to apply. Defaults to ``3``.

        :rtype: numeric or :obj:`None <python:None>`
        """
        return self._border_radius

    @border_radius.setter
    def border_radius(self, value):
        self._border_radius = validators.numeric(value, allow_empty = True)

    @property
    def color(self) -> Optional[str]:
        """The color to apply. Defaults to ``'#335cad'``.

        :rtype: :class:`str <python:str>` or :obj:`None <python:None>`
        """
        return self._color

    @color.setter
    def color(self, value):
        self._color = validators.string(value, allow_empty = True)

    @property
    def line_width(self) -> Optional[int | float | Decimal]:
        """The line width to apply. Defaults to ``2``.

        :rtype: numeric
        """
        return self._line_width

    @line_width.setter
    def line_width(self, value):
        self._line_width = validators.numeric(value, allow_empty = True)

    @classmethod
    def _get_kwargs_from_dict(cls, as_dict):
        kwargs = {
            'border_radius': as_dict.get('borderRadius', None),
            'color': as_dict.get('color', None),
            'line_width': as_dict.get('lineWidth', None)
        }
        return kwargs

    def _to_untrimmed_dict(self, in_cls = None) -> dict:
        untrimmed = {
            'borderRadius': self.border_radius,
            'color': self.color,
            'lineWidth': self.line_width
        }
        return untrimmed


class FocusBorder(HighchartsMeta):
    """Options for the focus border drawn around elements while navigating through them.
    """

    def __init__(self, **kwargs):
        self._enabled = None
        self._hide_browser_focus_outline = None
        self._margin = None
        self._style = None

        self.enabled = kwargs.get('enabled', None)
        self.hide_browser_focus_outline = kwargs.get('hide_browser_focus_outline', None)
        self.margin = kwargs.get('margin', None)
        self.style = kwargs.get('style', None)

    @property
    def enabled(self) -> Optional[bool]:
        """If ``True``, enables the focus border for the chart. Defaults to ``True``.

        :returns: Flag enabling or disabling the focus border around chart elements during
          navigation.
        :rtype: :class:`bool <python:bool>` or :obj:`None <python:None>`
        """
        return self._enabled

    @enabled.setter
    def enabled(self, value):
        if value is None:
            self._enabled = None
        else:
            self._enabled = bool(value)

    @property
    def hide_browser_focus_outline(self) -> Optional[bool]:
        """If ``True``, hides the default browser-generated focus border. Defaults to
        ``True``.

        :returns: Flag indicating whether to hide the default browser focus border.
        :rtype: :class:`bool <python:bool>` or :obj:`None <python:None>`
        """
        return self._hide_browser_focus_outline

    @hide_browser_focus_outline.setter
    def hide_browser_focus_outline(self, value):
        if value is None:
            self._hide_browser_focus_outline = None
        else:
            self._hide_browser_focus_outline = bool(value)

    @property
    def margin(self) -> Optional[int | float | Decimal]:
        """The focus border margin around the elements. Defaults to ``2``.

        :returns: The focus border margin around the elements.
        :rtype: numeric or :obj:`None <python:None>`
        """
        return self._margin

    @margin.setter
    def margin(self, value):
        self._margin = validators.numeric(value, allow_empty = True)

    @property
    def style(self) -> Optional[FocusBorderStyle]:
        """Style options for the focus border drawn around elements while navigating
        through them.

        .. warning::

          Some browsers, in addition, draw their own borders for focused elements.
          These automatic borders can not be styled by Highcharts.

        In styled mode, the border is given the ``.highcharts-focus-border`` class.

        :returns: Style options for the focus border.
        :rtype: :class:`FocusBorderStyle` or :obj:`None <python:None>`
        """
        return self._style

    @style.setter
    @class_sensitive(FocusBorderStyle)
    def style(self, value):
        self._style = value

    @classmethod
    def _get_kwargs_from_dict(cls, as_dict):
        kwargs = {
            'enabled': as_dict.get('enabled', None),
            'hide_browser_focus_outline': as_dict.get('hideBrowserFocusOutline', None),
            'margin': as_dict.get('margin', None),
            'style': as_dict.get('style', None)
        }
        return kwargs

    def _to_untrimmed_dict(self, in_cls = None) -> dict:
        untrimmed = {
            'enabled': self.enabled,
            'hideBrowserFocusOutline': self.hide_browser_focus_outline,
            'margin': self.margin,
            'style': self.style
        }
        return untrimmed



from typing import Optional
from decimal import Decimal

from validator_collection import validators, checkers

from highcharts_core import constants, errors, utility_functions
from highcharts_core.decorators import class_sensitive
from highcharts_core.metaclasses import HighchartsMeta, JavaScriptDict
from highcharts_core.utility_classes.gradients import Gradient
from highcharts_core.utility_classes.patterns import Pattern
from highcharts_core.utility_classes.events import PointEvents
from highcharts_core.options.series.data.accessibility import DataPointAccessibility


class DataCore(HighchartsMeta):
    """Primary base class for describing a data point."""

    def __init__(self, **kwargs):
        self._color = None
        self._events = None
        self._id = None
        self._label_rank = None
        self._name = None

        self.color = kwargs.get('color', None)
        self.events = kwargs.get('events', None)
        self.id = kwargs.get('id', None)
        self.label_rank = kwargs.get('label_rank', None) or kwargs.get('labelrank', None)
        self.name = kwargs.get('name', None)

    @property
    def color(self) -> Optional[str | Gradient | Pattern]:
        """The color of the individual data point. Defaults to :obj:`None <python:None>`.

        :rtype: :obj:`None <python:None>`, :class:`Gradient`, :class:`Pattern`, or
          :class:`str <python:str>`
        """
        return self._color

    @color.setter
    def color(self, value):
        self._color = utility_functions.validate_color(value)

    @property
    def events(self) -> Optional[PointEvents]:
        """Event handlers for individual data points.

        :rtype: :class:`PointEvents` or :obj:`None <python:None>`
        """
        return self._events

    @events.setter
    @class_sensitive(PointEvents)
    def events(self, value):
        self._events = value

    @property
    def id(self) -> Optional[str]:
        """The id of the data point. Defaults to :obj:`None <python:None>`.

        .. note::

          This can be used (in JavaScript) after render time to get a pointer to the point
          object through ``chart.get()``.

        :rtype: :class:`str <python:str>` or :obj:`None <python:None>`
        """
        return self._id

    @id.setter
    def id(self, value):
        self._id = validators.string(value, allow_empty = True)

    @property
    def label_rank(self) -> Optional[int | float | Decimal]:
        """The rank for this point's data label in the case of collision. Defaults to
        :obj:`None <python:None>`.

        .. note::

          If two data labels are about to overlap, the data label for the point with the
          highest ``label_rank`` will be shown.

        :rtype: numeric or :obj:`None <python:None>`
        """
        return self._label_rank

    @label_rank.setter
    def label_rank(self, value):
        self._label_rank = validators.numeric(value, allow_empty = True)

    @property
    def name(self) -> Optional[str]:
        """The name to display for the point in data labels, tooltips, in legends, etc.
        Defaults to :obj:`None <python:None>`.

        :rtype: :class:`str <python:str>` or :obj:`None <python:None>`
        """
        return self._name

    @name.setter
    def name(self, value):
        self._name = validators.string(value, allow_empty = True)

    @classmethod
    def _get_kwargs_from_dict(cls, as_dict):
        """Convenience method which returns the keyword arguments used to initialize the
        class from a Highcharts Javascript-compatible :class:`dict <python:dict>` object.

        :param as_dict: The HighCharts JS compatible :class:`dict <python:dict>`
          representation of the object.
        :type as_dict: :class:`dict <python:dict>`

        :returns: The keyword arguments that would be used to initialize an instance.
        :rtype: :class:`dict <python:dict>`

        """
        kwargs = {
            'color': as_dict.get('color', None),
            'events': as_dict.get('events', None),
            'id': as_dict.get('id', None),
            'label_rank': as_dict.get('labelRank',
                                      None) or as_dict.get('labelrank',
                                                           None),
            'name': as_dict.get('name', None),
        }

        return kwargs

    def _to_untrimmed_dict(self, in_cls = None) -> dict:
        untrimmed = {
            'color': self.color,
            'events': self.events,
            'id': self.id,
            'labelrank': self.label_rank,
            'name': self.name,
        }

        return untrimmed


class DataBase(DataCore):
    """Extended base class for describing a data point."""

    def __init__(self, **kwargs):
        self._accessibility = None
        self._class_name = None
        self._color_index = None
        self._custom = None
        self._description = None
        self._selected = None

        self.accessibility = kwargs.get('accessibility', None)
        self.class_name = kwargs.get('class_name', None)
        self.color_index = kwargs.get('color_index', None)
        self.custom = kwargs.get('custom', None)
        self.description = kwargs.get('description', None)
        self.selected = kwargs.get('selected', None)

        super().__init__(**kwargs)

    @property
    def accessibility(self) -> Optional[DataPointAccessibility]:
        """Accessibility options for a data point.

        :rtype: :class:`DataPointAccessibility` or :obj:`None <python:None>`
        """
        return self._accessibility

    @accessibility.setter
    @class_sensitive(DataPointAccessibility)
    def accessibility(self, value):
        self._accessibility = value

    @property
    def class_name(self) -> Optional[str]:
        """The additional CSS class name to apply to the data point's graphical elements.

        :rtype: :class:`str <python:str>` or :obj:`None <python:None>`
        """
        return self._class_name

    @class_name.setter
    def class_name(self, value):
        self._class_name = validators.string(value, allow_empty = True)

    @property
    def color_index(self) -> Optional[int]:
        """When operating in :term:`styled mode`, a specific color index to use for the
        point, so its graphic representations are given the class name
        ``highcharts-color-{n}``. Defaults to :obj:`None <python:None>`.

        .. tip::
        
          .. versionadded:: Highcharts (JS) v.11

          With Highcharts (JS) v.11, using CSS variables of the form ``--highcharts-color-{n}`` make
          changing the color scheme very simple.

        :rtype: :class:`int <python:int>` or :obj:`None <python:None>`
        """
        return self._color_index

    @color_index.setter
    def color_index(self, value):
        self._color_index = validators.integer(value,
                                               allow_empty = True,
                                               minimum = 0)

    @property
    def custom(self) -> Optional[JavaScriptDict]:
        """A reserved subspace to store options and values for customized functionality.

        Here you can add additional data for your own event callbacks and formatter
        callbacks.

        :rtype: :class:`dict <python:dict>` or :obj:`None <python:None>`
        """
        return self._custom

    @custom.setter
    @class_sensitive(JavaScriptDict)
    def custom(self, value):
        self._custom = value

    @property
    def description(self) -> Optional[str]:
        """A description of the data point to add to the screen reader information about
        the data point.

        :rtype: :class:`str <python:str>`
        """
        return self._description

    @description.setter
    def description(self, value):
        self._description = validators.string(value, allow_empty = True)

    @property
    def selected(self) -> Optional[bool]:
        """If ``True``, indicates that the data point is initially selected. Defaults to
        :obj:`None <python:None>`, which behaves as ``False``.

        :rtype: :class:`bool <python:bool>` or :obj:`None <python:None>`
        """
        return self._selected

    @selected.setter
    def selected(self, value):
        if value is None:
            self._selected = None
        else:
            self._selected = bool(value)

    @classmethod
    def _get_kwargs_from_dict(cls, as_dict):
        """Convenience method which returns the keyword arguments used to initialize the
        class from a Highcharts Javascript-compatible :class:`dict <python:dict>` object.

        :param as_dict: The HighCharts JS compatible :class:`dict <python:dict>`
          representation of the object.
        :type as_dict: :class:`dict <python:dict>`

        :returns: The keyword arguments that would be used to initialize an instance.
        :rtype: :class:`dict <python:dict>`

        """
        kwargs = {
            'accessibility': as_dict.get('accessibility', None),
            'class_name': as_dict.get('className', None),
            'color': as_dict.get('color', None),
            'color_index': as_dict.get('colorIndex', None),
            'custom': as_dict.get('custom', None),
            'description': as_dict.get('description', None),
            'events': as_dict.get('events', None),
            'id': as_dict.get('id', None),
            'label_rank': as_dict.get('labelRank',
                                      None) or as_dict.get('labelrank',
                                                           None),
            'name': as_dict.get('name', None),
            'selected': as_dict.get('selected', None),
        }

        return kwargs

    def _to_untrimmed_dict(self, in_cls = None) -> dict:
        untrimmed = {
            'accessibility': self.accessibility,
            'className': self.class_name,
            'color': self.color,
            'colorIndex': self.color_index,
            'custom': self.custom,
            'description': self.description,
            'events': self.events,
            'id': self.id,
            'labelrank': self.label_rank,
            'name': self.name,
            'selected': self.selected,
        }

        parent_as_dict = super()._to_untrimmed_dict(in_cls = in_cls)
        for key in parent_as_dict:
            untrimmed[key] = parent_as_dict[key]

        return untrimmed

    @classmethod
    def from_array(cls, value):
        """Creates a collection of data point instances, parsing the contents of ``value``
        as an array (iterable). This method is specifically used to parse data that is
        input to **Highcharts for Python** without property names, in an array-organized
        structure as described in the `Highcharts JS <https://www.highcharts.com>`__
        documentation.

          .. seealso::

            The specific structure of the expected array is highly dependent on the type
            of data point that the series needs, which itself is dependent on the series
            type itself.

            Please review the detailed :ref:`series documentation <series_documentation>`
            for series type-specific details of relevant array structures.

          .. note::

            An example of how this works for a simple
            :class:`LineSeries <highcharts_core.options.series.area.LineSeries>` (which
            uses
            :class:`CartesianData <highcharts_core.options.series.data.cartesian.CartesianData>`
            data points) would be:

            .. code-block:: python

              my_series = LineSeries()

              # A simple array of numerical values which correspond to the Y value of the
              # data point
              my_series.data = [0, 5, 3, 5]

              # An array containing 2-member arrays (corresponding to the X and Y values
              # of the data point)
              my_series.data = [
                  [0, 0],
                  [1, 5],
                  [2, 3],
                  [3, 5]
              ]

              # An array of dict with named values
              my_series.data = [
                {
                    'x': 0,
                    'y': 0,
                    'name': 'Point1',
                    'color': '#00FF00'
                },
                {
                    'x': 1,
                    'y': 5,
                    'name': 'Point2',
                    'color': '#CCC'
                },
                {
                    'x': 2,
                    'y': 3,
                    'name': 'Point3',
                    'color': '#999'
                },
                {
                    'x': 3,
                    'y': 5,
                    'name': 'Point4',
                    'color': '#000'
                }
              ]

        :param value: The value that should contain the data which will be converted into
          data point instances.

          .. note::

            If ``value`` is not an iterable, it will be converted into an iterable to be
            further de-serialized correctly.

        :type value: iterable

        :returns: Collection of :term:`data point` instances (descended from
          :class:`DataBase <highcharts_core.options.series.data.base.DataBase>`)
        :rtype: :class:`list <python:list>` of
          :class:`DataBase <highcharts_core.options.series.data.base.DataBase>`
          descendant instances
        """
        if not value:
            return []
        elif checkers.is_string(value):
            try:
                value = validators.json(value)
            except (ValueError, TypeError):
                pass
        elif not checkers.is_iterable(value):
            value = [value]

        collection = []
        for item in value:
            if checkers.is_type(item, 'DataBase'):
                as_obj = item
            elif checkers.is_dict(item):
                as_obj = cls.from_dict(item)
            elif item is None or isinstance(item, constants.EnforcedNullType):
                as_obj = cls()
            else:
                raise errors.HighchartsValueError(f'each data point supplied must either '
                                                  f'be a DataBase Data Point or be '
                                                  f'coercable to one. Could not coerce: '
                                                  f'{item}')
            collection.append(as_obj)

        return collection
